import json
from pathlib import Path

from rdflib import ConjunctiveGraph, Dataset, Graph


def test_hext_graph():
    """Tests single-grant (not context-aware) data"""
    g = Graph()
    turtle_data = """
            PREFIX ex: <http://example.com/>
            PREFIX owl: <http://www.w3.org/2002/07/owl#>
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>

            ex:s1
                ex:p1 ex:o1 , ex:o2 ;
                ex:p2 [
                    a owl:Thing ;
                    rdf:value "thingy" ;
                ] ;
                ex:p3 "Object 3" , "Object 4 - English"@en ;
                ex:p4 "2021-12-03"^^xsd:date ;
                ex:p5 42 ;
                ex:p6 "42" ;
                ex:p7 true ;
                ex:p8 "false"^^xsd:boolean ;
            .
            """

    g.parse(data=turtle_data, format="turtle")
    out = g.serialize(format="hext")
    # note: can't test for BNs in result as they will be different every time
    testing_lines = [
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o2", "globalId", "", ""]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p3", "Object 3", "http://www.w3.org/2001/XMLSchema#string", "", ""]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p3", "Object 4 - English", "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString", "en", ""]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o1", "globalId", "", ""]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p4", "2021-12-03", "http://www.w3.org/2001/XMLSchema#date", "", ""]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p6", "42", "http://www.w3.org/2001/XMLSchema#string", "", ""]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p7", "true", "http://www.w3.org/2001/XMLSchema#boolean", "", ""]',
        ],
        [
            False,
            '"http://www.w3.org/1999/02/22-rdf-syntax-ns#value", "thingy", "http://www.w3.org/2001/XMLSchema#string", "", ""]',
        ],
        [False, '["http://example.com/s1", "http://example.com/p2"'],
        [
            False,
            '["http://example.com/s1", "http://example.com/p5", "42", "http://www.w3.org/2001/XMLSchema#integer", "", ""]',
        ],
        [
            False,
            '"http://www.w3.org/1999/02/22-rdf-syntax-ns#type", "http://www.w3.org/2002/07/owl#Thing", "globalId", "", ""]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p8", "false", "http://www.w3.org/2001/XMLSchema#boolean", "", ""]',
        ],
    ]
    for line in out.splitlines():
        normalized_line = line.replace(", ", ",").strip()
        for test in testing_lines:
            normalized_test = test[1].replace(", ", ",").strip()
            if normalized_test in normalized_line:
                test[0] = True

    assert all([x[0] for x in testing_lines])


def test_hext_cg():
    """Tests ConjunctiveGraph data"""
    d = ConjunctiveGraph()
    trig_data = """
            PREFIX ex: <http://example.com/>
            PREFIX owl: <http://www.w3.org/2002/07/owl#>
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>

            ex:g1 {
                ex:s1
                    ex:p1 ex:o1 , ex:o2 ;
                    ex:p2 [
                        a owl:Thing ;
                        rdf:value "thingy" ;
                    ] ;
                    ex:p3 "Object 3" , "Object 4 - English"@en ;
                    ex:p4 "2021-12-03"^^xsd:date ;
                    ex:p5 42 ;
                    ex:p6 "42" ;
                .
            }

            ex:g2 {
                ex:s1
                    ex:p1 ex:o1 , ex:o2 ;
                .
                ex:s11 ex:p11 ex:o11 , ex:o12 .
            }

            # default graph triples
            ex:s1 ex:p1 ex:o1 , ex:o2 .
            ex:s21 ex:p21 ex:o21 , ex:o22 .

            # other default graph triples
            {
                ex:s1 ex:p1 ex:o1 , ex:o2 .
            }
           """
    d.parse(data=trig_data, format="trig", publicID=d.default_context.identifier)
    out = d.serialize(format="hext")
    # note: cant' test for BNs in result as they will be different every time
    testing_lines = [
        [
            False,
            '["http://example.com/s21", "http://example.com/p21", "http://example.com/o21", "globalId", "", ""]',
        ],
        [
            False,
            '["http://example.com/s21", "http://example.com/p21", "http://example.com/o22", "globalId", "", ""]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o2", "globalId", "", ""]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o1", "globalId", "", ""]',
        ],
        [
            False,
            '["http://example.com/s11", "http://example.com/p11", "http://example.com/o12", "globalId", "", "http://example.com/g2"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o2", "globalId", "", "http://example.com/g2"]',
        ],
        [
            False,
            '["http://example.com/s11", "http://example.com/p11", "http://example.com/o11", "globalId", "", "http://example.com/g2"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o1", "globalId", "", "http://example.com/g2"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o2", "globalId", "", "http://example.com/g1"]',
        ],
        [False, '["http://example.com/s1", "http://example.com/p2"'],
        [
            False,
            '"http://www.w3.org/1999/02/22-rdf-syntax-ns#value", "thingy", "http://www.w3.org/2001/XMLSchema#string", "", "http://example.com/g1"]',
        ],
        [
            False,
            '"http://www.w3.org/1999/02/22-rdf-syntax-ns#type", "http://www.w3.org/2002/07/owl#Thing", "globalId", "", "http://example.com/g1"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p3", "Object 4 - English", "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString", "en", "http://example.com/g1"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p6", "42", "http://www.w3.org/2001/XMLSchema#string", "", "http://example.com/g1"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p4", "2021-12-03", "http://www.w3.org/2001/XMLSchema#date", "", "http://example.com/g1"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o1", "globalId", "", "http://example.com/g1"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p5", "42", "http://www.w3.org/2001/XMLSchema#integer", "", "http://example.com/g1"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p3", "Object 3", "http://www.w3.org/2001/XMLSchema#string", "", "http://example.com/g1"]',
        ],
    ]
    for line in out.splitlines():
        normalized_line = line.replace(", ", ",").strip()
        for test in testing_lines:
            normalized_test = test[1].replace(", ", ",").strip()
            if normalized_test in normalized_line:
                test[0] = True

    assert all([x[0] for x in testing_lines])


def test_hext_dataset():
    """Tests context-aware (multigraph) data"""
    d = Dataset()
    trig_data = """
            PREFIX ex: <http://example.com/>
            PREFIX owl: <http://www.w3.org/2002/07/owl#>
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>

            ex:g1 {
                ex:s1
                    ex:p1 ex:o1 , ex:o2 ;
                    ex:p2 [
                        a owl:Thing ;
                        rdf:value "thingy" ;
                    ] ;
                    ex:p3 "Object 3" , "Object 4 - English"@en ;
                    ex:p4 "2021-12-03"^^xsd:date ;
                    ex:p5 42 ;
                    ex:p6 "42" ;
                .
            }

            ex:g2 {
                ex:s1
                    ex:p1 ex:o1 , ex:o2 ;
                .
                ex:s11 ex:p11 ex:o11 , ex:o12 .
            }

            # default graph triples
            ex:s1 ex:p1 ex:o1 , ex:o2 .
            ex:s21 ex:p21 ex:o21 , ex:o22 .
           """
    d.parse(data=trig_data, format="trig", publicID=d.default_context.identifier)
    out = d.serialize(format="hext")
    # note: cant' test for BNs in result as they will be different ever time
    testing_lines = [
        [
            False,
            '["http://example.com/s21", "http://example.com/p21", "http://example.com/o21", "globalId", "", ""]',
        ],
        [
            False,
            '["http://example.com/s21", "http://example.com/p21", "http://example.com/o22", "globalId", "", ""]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o2", "globalId", "", ""]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o1", "globalId", "", ""]',
        ],
        [
            False,
            '["http://example.com/s11", "http://example.com/p11", "http://example.com/o12", "globalId", "", "http://example.com/g2"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o2", "globalId", "", "http://example.com/g2"]',
        ],
        [
            False,
            '["http://example.com/s11", "http://example.com/p11", "http://example.com/o11", "globalId", "", "http://example.com/g2"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o1", "globalId", "", "http://example.com/g2"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o2", "globalId", "", "http://example.com/g1"]',
        ],
        [False, '["http://example.com/s1", "http://example.com/p2"'],
        [
            False,
            '"http://www.w3.org/1999/02/22-rdf-syntax-ns#value", "thingy", "http://www.w3.org/2001/XMLSchema#string", "", "http://example.com/g1"]',
        ],
        [
            False,
            '"http://www.w3.org/1999/02/22-rdf-syntax-ns#type", "http://www.w3.org/2002/07/owl#Thing", "globalId", "", "http://example.com/g1"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p3", "Object 4 - English", "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString", "en", "http://example.com/g1"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p6", "42", "http://www.w3.org/2001/XMLSchema#string", "", "http://example.com/g1"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p4", "2021-12-03", "http://www.w3.org/2001/XMLSchema#date", "", "http://example.com/g1"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p1", "http://example.com/o1", "globalId", "", "http://example.com/g1"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p5", "42", "http://www.w3.org/2001/XMLSchema#integer", "", "http://example.com/g1"]',
        ],
        [
            False,
            '["http://example.com/s1", "http://example.com/p3", "Object 3", "http://www.w3.org/2001/XMLSchema#string", "", "http://example.com/g1"]',
        ],
    ]
    for line in out.splitlines():
        normalized_line = line.replace(", ", ",").strip()
        for test in testing_lines:
            normalized_test = test[1].replace(", ", ",").strip()
            if normalized_test in normalized_line:
                test[0] = True

    assert all([x[0] for x in testing_lines])


def test_hext_json_representation():
    """Tests to see if every link in the ND-JSON Hextuple result is, in fact, JSON"""
    d = Dataset()
    trig_data = """
            PREFIX ex: <http://example.com/>
            PREFIX owl: <http://www.w3.org/2002/07/owl#>
            PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
            PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>

            ex:g1 {
                ex:s1
                    ex:p1 ex:o1 , ex:o2 ;
                    ex:p2 [
                        a owl:Thing ;
                        rdf:value "thingy" ;
                    ] ;
                    ex:p3 "Object 3" , "Object 4 - English"@en ;
                    ex:p4 "2021-12-03"^^xsd:date ;
                    ex:p5 42 ;
                    ex:p6 "42" ;
                .
            }

            ex:g2 {
                ex:s1
                    ex:p1 ex:o1 , ex:o2 ;
                .
                ex:s11 ex:p11 ex:o11 , ex:o12 .
            }

            # default graph triples
            ex:s1 ex:p1 ex:o1 , ex:o2 .
            ex:s21 ex:p21 ex:o21 , ex:o22 .
           """
    d.parse(data=trig_data, format="trig")
    out = d.serialize(format="hext")
    for line in out.splitlines():
        j = json.loads(line)
        assert isinstance(j, list)


def test_hext_dataset_linecount():
    d = Dataset()
    assert len(d) == 0
    d.parse(
        Path(__file__).parent.parent / "data/test_parser_hext_multigraph.ndjson",
        format="hext",
        publicID=d.default_context.identifier,
    )
    total_triples = 0
    # count all the triples in the Dataset
    for context in d.contexts():
        for triple in context.triples((None, None, None)):
            total_triples += 1
    assert total_triples == 18

    # count the number of serialized Hextuples, should be 22, as per the original file
    lc = len(d.serialize(format="hext").splitlines())
    assert lc == 22


def test_roundtrip():
    d = Dataset()
    d.parse(
        Path(__file__).parent.parent / "data/test_parser_hext_multigraph.ndjson",
        format="hext",
        publicID=d.default_context.identifier,
    )
    d.default_union = True
    with open(
        str(Path(__file__).parent.parent / "data/test_parser_hext_multigraph.ndjson")
    ) as i:
        ordered_input = "".join(sorted(i.readlines())).strip()
        normalized_ordered_input = ordered_input.replace(", ", ",")

    ordered_output = "\n".join(sorted(d.serialize(format="hext").split("\n"))).strip()
    normalized_ordered_output = ordered_output.replace(", ", ",")

    assert normalized_ordered_output == normalized_ordered_input


# def _make_large_graph():
#     import random
#
#     EX = Namespace("http://example.com/")
#     g = Graph()
#
#     for i in range(1000):
#         s = EX["s" + str(random.randint(1, 10000)).zfill(5)]
#         p = EX["p" + str(random.randint(1, 10000)).zfill(5)]
#         o_r = random.randint(1, 10000)
#         if o_r > 5000:
#             o = EX["p" + str(o_r).zfill(5)]
#         else:
#             o = Literal("p" + str(o_r).zfill(5))
#         g.add((s, p, o))
#     return g
#
#
# def test_hext_scaling():
#     g = _make_large_graph()
#     g.serialize(destination="large.ndjson", format="hext")
#
#
if __name__ == "__main__":
    test_roundtrip()



import pytest

from rdflib import OWL, RDF, RDFS, XSD, Graph, Literal, Namespace, URIRef, logger
from rdflib.extras.infixowl import (
    AllClasses,
    AllProperties,
    BooleanClass,
    Class,
    ClassNamespaceFactory,
    CommonNSBindings,
    ComponentTerms,
    DeepClassClear,
    EnumeratedClass,
    GetIdentifiedClasses,
    Individual,
    Infix,
    MalformedClassError,
    Property,
    Restriction,
    classOrTerm,
    exactly,
    generateQName,
    max,
    min,
    only,
    some,
    value,
)
from test.data import CONTEXT0, TEST_DATA_DIR

EXNS = Namespace("http://example.org/vocab/")
PZNS = Namespace(
    "http://www.co-ode.org/ontologies/pizza/2005/10/18/classified/pizza.owl#"
)


def test_infix_operators_dunders():
    def fn(other, arg=""):
        return None

    i = Infix(fn)

    assert i.__rlshift__("foo") is not None
    assert i.__rshift__("foo") is None
    assert i.__rmatmul__("foo") is not None
    assert i.__matmul__("foo") is None
    assert i.__call__("foo", "baz") is None


def test_infix_operators():
    g = Graph()
    g.bind("ex", EXNS)

    # Broken since forever
    with pytest.raises(Exception, match="Only URIRefs or Paths can be in paths!"):
        EXNS.hasParent | some | Class(EXNS.Physician, graph=g)

    # Will break when RDF Star syntax is added
    res = EXNS.hasParent << some >> Class(EXNS.Physician, graph=g)
    assert str(res) == "( ex:hasParent SOME ex:Physician )"

    res = EXNS.hasParent @ some @ Class(EXNS.Physician, graph=g)
    assert str(res) == "( ex:hasParent SOME ex:Physician )"


def test_generateqname():
    g = Graph()
    g.bind("ex", EXNS)
    assert (
        generateQName(g, URIRef("http://example.org/vocab/ontology")) == "ex:ontology"
    )


def test_classorterm():
    term = URIRef("http://example.org/vocab/D")
    assert classOrTerm(Class(EXNS.D)) == term
    assert classOrTerm(term) == term


def test_getidentifiedclasses():
    g = Graph()
    g.bind("ex", PZNS)
    g.parse(TEST_DATA_DIR / "owl" / "pizza.owl", format="xml")
    assert len(list(GetIdentifiedClasses(g))) == 97


class Foo:
    pass


def test_common_ns_bindings():
    g = Graph(bind_namespaces=None)
    CommonNSBindings(g, additionalNS={"ex": EXNS})


def test_allclasses():
    g = Graph()
    g.bind("ex", PZNS)

    g.parse(TEST_DATA_DIR / "owl" / "pizza.owl", format="xml")

    for c in AllClasses(g):
        assert isinstance(c, Class)


def test_check_allclasses():
    from test.data import BOB, MICHEL, TAREK

    g = Graph()
    g.bind("ex", PZNS)

    g.add((TAREK, RDF.type, OWL.Class))
    g.add((MICHEL, RDF.type, OWL.Class))
    g.add((BOB, RDF.type, OWL.Class))

    assert set(g.subjects(predicate=RDF.type, object=OWL.Class)) == {
        URIRef("urn:example:bob"),
        URIRef("urn:example:michel"),
        URIRef("urn:example:tarek"),
    }


def test_check_allproperties():
    g = Graph()
    g.bind("ex", PZNS)

    g.parse(TEST_DATA_DIR / "owl" / "pizza.owl", format="xml")

    for p in AllProperties(g):
        assert isinstance(p, Property)


EXCL = ClassNamespaceFactory(EXNS)


def test_classnamespacefactory():
    leg = EXCL.Leg

    x = EXCL.__getitem__(leg.identifier)

    assert x is not None


def test_componentterms():
    g = Graph()
    g.bind("ex", PZNS)

    Individual.factoryGraph = g

    g.parse(TEST_DATA_DIR / "owl" / "pizza.owl", format="xml")

    owlcls = list(AllClasses(g))[-1]

    assert isinstance(owlcls, Class)

    for c in ComponentTerms(owlcls):
        assert isinstance(c, Class)


def test_componentterms_extended():
    g = Graph()
    g.bind("ex", PZNS)
    Individual.factoryGraph = g

    isPartOf = Property(EXNS.isPartOf)  # noqa: N806
    g.add((isPartOf.identifier, RDF.type, OWL.TransitiveProperty))

    hasLocation = Property(EXNS.hasLocation, subPropertyOf=[isPartOf])  # noqa: N806
    g.add((hasLocation.identifier, RDFS.subPropertyOf, isPartOf.identifier))

    leg = EXCL.Leg
    knee = EXCL.Knee
    joint = EXCL.Joint
    kneeJoint = EXCL.KneeJoint  # noqa: N806
    legStructure = EXCL.LegStructure  # noqa: N806

    kneeJoint.equivalentClass = [joint & (isPartOf @ some @ knee)]

    structure = EXCL.Structure

    legStructure.equivalentClass = [structure & (isPartOf @ some @ leg)]

    structure += leg
    structure += joint

    locatedInLeg = hasLocation @ some @ leg  # noqa: N806
    locatedInLeg += knee  # noqa: N806

    for c in ComponentTerms(knee):
        assert isinstance(c, Class)

    for c in ComponentTerms(legStructure):
        assert isinstance(c, Class)

    logger.debug(f"G\n{g.serialize(format='ttl')}")


def test_raise_malformedclasserror():
    e = MalformedClassError("Malformed owl:Restriction")

    assert str(repr(e)) == "Malformed owl:Restriction"

    with pytest.raises(MalformedClassError, match="Malformed owl:Restriction"):
        raise e


def test_owlrdfproxylist():
    g = Graph()
    g.bind("ex", EXNS)
    Individual.factoryGraph = g

    ogbujiBros = EnumeratedClass(  # noqa: N806
        EXNS.ogbujiBros, members=[EXNS.chime, EXNS.uche, EXNS.ejike]
    )

    ogbujiSiblings = EnumeratedClass(  # noqa: N806
        EXNS.ogbujiBros, members=[EXNS.uche, EXNS.chime, EXNS.ejike]
    )

    assert ogbujiBros == ogbujiSiblings

    assert EXNS.chime in ogbujiBros

    fred = EXNS.fred

    bert = EXNS.bert

    assert fred not in ogbujiBros

    ogbujiBros.append(fred)

    ogbujiBros += bert  # noqa: N806

    assert ogbujiBros[1] == EXNS.uche

    assert len(ogbujiBros) == 5

    del ogbujiBros[4]

    assert len(ogbujiBros) == 4

    ogbujiBros[4] = fred

    assert len(ogbujiBros) == 5

    male = Class(EXNS.Male, graph=g)
    female = Class(EXNS.Female, graph=g)
    human = Class(EXNS.Human, graph=g)

    youngPerson = Class(EXNS.YoungPerson, graph=g)  # noqa: N806

    youngWoman = female & human & youngPerson  # noqa: N806
    assert isinstance(youngWoman, BooleanClass)

    res = ogbujiBros.__eq__(youngWoman)
    assert res is False

    youngMan = male & human & youngPerson  # noqa: N806

    res = youngMan.__eq__(youngWoman)
    assert res is False


def test_deepclassclear():
    g = Graph()
    g.bind("ex", EXNS)
    Individual.factoryGraph = g

    classB = Class(EXNS.B)  # noqa: N806
    classC = Class(EXNS.C)  # noqa: N806
    classD = Class(EXNS.D)  # noqa: N806
    classE = Class(EXNS.E)  # noqa: N806
    classF = Class(EXNS.F)  # noqa: N806

    anonClass = EXNS.someProp << some >> classD  # noqa: N806

    assert str(anonClass) == "( ex:someProp SOME ex:D )"
    assert str(classF) == "Class: ex:F "

    classF += anonClass  # noqa: N806

    assert str(list(anonClass.subClassOf)) == "[Class: ex:F ]"

    classA = classE | classF | anonClass  # noqa: N806
    classG = Class(EXNS.G, complementOf=classA)  # noqa: N806
    classB += classA  # noqa: N806
    classA.equivalentClass = [Class()]
    classA.complementOf = classG
    classB.subClassOf = [EXNS.someProp << some >> classC]

    assert str(classA) == "( ex:E OR ex:F OR ( ex:someProp SOME ex:D ) )"

    DeepClassClear(classA)
    assert str(classA) == "(  )"

    assert list(anonClass.subClassOf) == []

    assert str(classB) == "Class: ex:B SubClassOf: ( ex:someProp SOME ex:C )"

    otherClass = classD | anonClass  # noqa: N806

    assert str(otherClass) == "( ex:D OR ( ex:someProp SOME ex:D ) )"

    DeepClassClear(otherClass)
    assert str(otherClass) == "(  )"

    otherClass.delete()
    assert list(g.triples((otherClass.identifier, None, None))) == []


def test_booleanclassextenthelper():
    g = Graph()
    g.bind("ex", EXNS)

    Individual.factoryGraph = g

    fire = Class(EXNS.Fire)
    water = Class(EXNS.Water)

    tc1 = BooleanClass(members=[fire, water])
    assert str(tc1) == "( ex:Fire AND ex:Water )"

    tc2 = BooleanClass(operator=OWL.unionOf, members=[fire, water])
    assert str(tc2) == "( ex:Fire OR ex:Water )"


def test_changeoperator():
    # Converts a unionOf / intersectionOf class expression into one
    # that instead uses the given operator

    g = Graph(identifier=CONTEXT0)
    g.bind("ex", EXNS)
    Individual.factoryGraph = g

    fire = Class(EXNS.Fire)
    water = Class(EXNS.Water)

    testclass = BooleanClass(members=[fire, water])

    assert str(testclass) == "( ex:Fire AND ex:Water )"

    testclass.changeOperator(OWL.unionOf)
    assert repr(testclass) == "( ex:Fire OR ex:Water )"

    try:
        testclass.changeOperator(OWL.unionOf)
    except Exception as e:
        assert str(e) == "The new operator is already being used!"


def test_cardinality_zero():
    g = Graph()
    g.bind("ex", EXNS)

    prop = Property(EXNS.someProp, baseType=OWL.DatatypeProperty)
    r = Restriction(
        prop, graph=g, cardinality=Literal(0, datatype=XSD.nonNegativeInteger)
    )

    assert str(r) == "( ex:someProp EQUALS 0 )"


def test_textual_infix_operators():
    g = Graph()
    g.bind("ex", EXNS)
    Individual.factoryGraph = g

    isPartOf = Property(EXNS.isPartOf)  # noqa: N806
    g.add((isPartOf.identifier, RDF.type, OWL.TransitiveProperty))

    knee = EXCL.Knee

    x = isPartOf @ some @ knee

    assert x is not None

    y = isPartOf @ only @ knee

    assert y is not None

    y = isPartOf @ exactly @ Literal("1", datatype=XSD.integer)

    y = isPartOf @ min @ Literal("1", datatype=XSD.integer)

    y = isPartOf @ max @ Literal("1", datatype=XSD.integer)

    y = isPartOf @ value @ Literal("1", datatype=XSD.integer)



"""
SPARQL Query using :meth:`rdflib.graph.Graph.query`

The method returns a :class:`~rdflib.query.Result`, iterating over
this yields :class:`~rdflib.query.ResultRow` objects

The variable bindings can be accessed as attributes of the row objects
For variable names that are not valid python identifiers, dict access
(i.e. with ``row[var] / __getitem__``) is also possible.

:attr:`~rdflib.query.Result.vars` contains the variables
"""

import logging
import sys
from pathlib import Path

import rdflib

EXAMPLES_DIR = Path(__file__).parent

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG, stream=sys.stderr)

    g = rdflib.Graph()
    g.parse(f"{EXAMPLES_DIR / 'foaf.n3'}", format="n3")

    # The QueryProcessor knows the FOAF prefix from the graph
    # which in turn knows it from reading the N3 RDF file
    for row in g.query("SELECT ?s WHERE { [] foaf:knows ?s .}"):
        # For select queries, the Result object is an iterable of ResultRow
        # objects.
        assert isinstance(row, rdflib.query.ResultRow)
        print(row.s)
        # or row["s"]
        # or row[rdflib.Variable("s")]



from __future__ import annotations

import itertools
import logging
from typing import Dict, Iterable, List, Optional, Set, Tuple

import pytest
from _pytest.mark.structures import ParameterSet

from rdflib.graph import Graph
from test.utils import GraphHelper
from test.utils.http import MethodName, MockHTTPResponse
from test.utils.httpservermock import ServedBaseHTTPServerMock


def make_test_query_construct_format_cases() -> Iterable[ParameterSet]:
    """
    This only tests with application/rdf+xml as other result formats are not
    supported.
    """
    graphs: List[Tuple[str, Graph]] = [
        (
            "basic",
            Graph().parse(
                format="turtle",
                data="""
        @prefix egdc: <http://example.com/> .

        egdc:s0 egdc:p0 egdc:o0;
            egdc:p1 egdc:o1;
            egdc:p2 "o2";
            egdc:p3 _:o3.


        egdc:s1 egdc:p4 egdc:o4;
            egdc:p5 egdc:o5;
            egdc:p6 "o6";
            egdc:p7 _:o7;
            egdc:p8 _:03.
        """,
            ),
        )
    ]
    response_format_encodings: List[Tuple[str, str, Set[Optional[str]]]] = [
        (
            "application/rdf+xml",
            "utf-8",
            {
                "application/rdf+xml",
                "application/rdf+xml;charset=utf-8",
                "application/rdf+xml; charset=utf-8",
                "application/rdf+xml; charset=UTF-8",
            },
        ),
    ]
    for (mime_type, encoding, content_types), (
        graph_label,
        expected_graph,
    ) in itertools.product(response_format_encodings, graphs):
        response_body = expected_graph.serialize(format=mime_type, encoding=encoding)

        for content_type in content_types:
            if content_type is None:
                content_type = f"{mime_type};charset={encoding}"
            response_headers: Dict[str, List[str]] = {"Content-Type": [content_type]}
            yield pytest.param(
                expected_graph,
                response_body,
                response_headers,
                id=f"{mime_type}-{encoding}-{graph_label}-ContentType_{content_type}",
            )


@pytest.mark.parametrize(
    [
        "expected_graph",
        "response_body",
        "response_headers",
    ],
    make_test_query_construct_format_cases(),
)
def test_query_construct_format(
    function_httpmock: ServedBaseHTTPServerMock,
    expected_graph: Graph,
    response_body: bytes,
    response_headers: Dict[str, List[str]],
) -> None:
    """
    This tests that bindings for a variable named var
    """
    logging.debug("response_headers = %s", response_headers)
    graph = Graph(
        # store_factory(mime_type),
        "SPARQLStore",
        identifier="http://example.com",
        bind_namespaces="none",
    )
    url = f"{function_httpmock.url}/query"
    logging.debug("opening %s", url)
    graph.open(url)
    query = """
    CONSTRUCT {
        ?s ?p ?o
    }
    WHERE {
        ?s ?p ?o
    }
    """

    function_httpmock.responses[MethodName.GET].append(
        MockHTTPResponse(
            200,
            "OK",
            response_body,
            response_headers,
        )
    )

    logging.debug("sending query %s", query)
    result = graph.query(query)
    logging.debug("result = %s", result)
    assert result.type == "CONSTRUCT"
    assert result.graph is not None
    GraphHelper.assert_isomorphic(expected_graph, result.graph)

    request = function_httpmock.requests[MethodName.GET].pop()
    logging.debug("request = %s", request)
    logging.debug("request.headers = %s", request.headers.as_string())
    assert request.path_query["query"][0] == query



from rdflib import Graph
from rdflib.namespace import RDFS
from rdflib.resource import Resource


def test_properties(rdfs_graph: Graph) -> None:
    """
    The properties of a `rdflib.resource.Resource` work as expected.
    """
    cres = Resource(rdfs_graph, RDFS.Container)
    assert cres.graph is rdfs_graph
    assert cres.identifier == RDFS.Container



import io

import rdflib
from rdflib import Graph
from test.utils import eq_


def test_datetime_datetime_subs_issue():
    """
    Test for query mentioned in the Issue #629
    https://github.com/RDFLib/rdflib/issues/629

    """

    data1 = """
    @prefix : <#>.
    @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#>.

    :C a rdfs:Class.

    :start a rdf:Property;
        rdfs:domain :C;
        rdfs:range xsd:dateTime.

    :end a rdf:Property;
        rdfs:domain :C;
        rdfs:range xsd:dateTime.

    :c1 a :C;
        :start "2016-01-01T20:00:00"^^xsd:dateTime;
        :end "2016-01-02T20:01:00"^^xsd:dateTime.

    :c2 a :C;
        :start "2016-01-01T20:05:00"^^xsd:dateTime;
        :end "2016-01-01T20:30:00"^^xsd:dateTime.
    """

    graph1 = Graph()
    f = io.StringIO(data1)
    graph1.parse(f, format="n3")

    result = graph1.query(
        """
    SELECT ?c ?duration
    WHERE {
        ?c :start ?start;
            :end ?end.
        BIND(?end - ?start AS ?duration)
    }
    """
    )

    answer = list(result)
    answer = sorted(answer)
    # expected result will be a list of 2 tuples
    # FirstElement of these tuples will be a node with a path of directory of saved project
    # Second Element while represent the actual durations

    expected_first_duration = rdflib.term.Literal(
        "P1DT1M",
        datatype=rdflib.term.URIRef("http://www.w3.org/2001/XMLSchema#duration"),
    )
    expected_second_duration = rdflib.term.Literal(
        "PT25M",
        datatype=rdflib.term.URIRef("http://www.w3.org/2001/XMLSchema#duration"),
    )

    eq_(answer[0][1], expected_first_duration)
    eq_(answer[1][1], expected_second_duration)


def test_datetime_duration_subs():
    """
    Test cases for subtraction operation between dateTime and duration

    """
    data = """
    @prefix : <#>.
    @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#>.
    """
    graph = Graph()
    f = io.StringIO(data)
    graph.parse(f, format="n3")

    # 1st Test Case

    result1 = graph.query(
        """
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT (?d - ?duration AS ?next_year)
    WHERE {
        VALUES (?duration ?d) {
            ("P1Y"^^xsd:yearMonthDuration "2019-05-28T12:14:45Z"^^xsd:dateTime)
            ("P1Y"^^xsd:yearMonthDuration "2019-05-28"^^xsd:date)
        }
    }
    """
    )
    expected = []
    expected.append(
        rdflib.term.Literal(
            "2018-05-28T12:14:45+00:00",
            datatype=rdflib.term.URIRef("http://www.w3.org/2001/XMLSchema#dateTime"),
        )
    )
    expected.append(
        rdflib.term.Literal(
            "2018-05-28",
            datatype=rdflib.term.URIRef("http://www.w3.org/2001/XMLSchema#date"),
        )
    )

    eq_(list(result1)[0][0], expected[0])
    eq_(list(result1)[1][0], expected[1])

    # 2nd Test Case

    result2 = graph.query(
        """
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT (?d - ?duration AS ?next_year)
    WHERE {
        VALUES (?duration ?d) {
            ("P3DT1H15M"^^xsd:dayTimeDuration "2000-10-30T11:12:00"^^xsd:dateTime)
            ("P3DT1H15M"^^xsd:dayTimeDuration "2000-10-30"^^xsd:date)
        }
    }
    """
    )

    expected = []
    expected.append(
        rdflib.term.Literal(
            "2000-10-27T09:57:00",
            datatype=rdflib.term.URIRef("http://www.w3.org/2001/XMLSchema#dateTime"),
        )
    )
    expected.append(
        rdflib.term.Literal(
            "2000-10-27",
            datatype=rdflib.term.URIRef("http://www.w3.org/2001/XMLSchema#date"),
        )
    )

    eq_(list(result2)[0][0], expected[0])
    eq_(list(result2)[1][0], expected[1])


def test_datetime_duration_add():
    """
    Test cases for addition operation between dateTime and duration

    """
    data = """
    @prefix : <#>.
    @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#>.
    """
    graph = Graph()
    f = io.StringIO(data)
    graph.parse(f, format="n3")

    # 1st Test case

    result1 = graph.query(
        """
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT (?d + ?duration AS ?next_year)
    WHERE {
        VALUES (?duration ?d) {
            ("P1Y"^^xsd:yearMonthDuration"2019-05-28T12:14:45Z"^^xsd:dateTime)
            ("P1Y"^^xsd:yearMonthDuration"2019-05-28"^^xsd:date)
        }
    }
    """
    )

    # print(list(result1))
    expected = []
    expected.append(
        rdflib.term.Literal(
            "2020-05-28T12:14:45+00:00",
            datatype=rdflib.term.URIRef("http://www.w3.org/2001/XMLSchema#dateTime"),
        )
    )
    expected.append(
        rdflib.term.Literal(
            "2020-05-28",
            datatype=rdflib.term.URIRef("http://www.w3.org/2001/XMLSchema#date"),
        )
    )

    eq_(list(result1)[0][0], expected[0])
    eq_(list(result1)[1][0], expected[1])

    # 2nd Test case

    result2 = graph.query(
        """
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT (?d + ?duration AS ?next_year)
    WHERE {
        VALUES (?duration ?d) {
            ("P3DT1H15M"^^xsd:dayTimeDuration "2000-10-30T11:12:00"^^xsd:dateTime)
            ("P3DT1H15M"^^xsd:dayTimeDuration "2000-10-30"^^xsd:date)
        }
    }
    """
    )

    # print(list(result2))
    expected = []
    expected.append(
        rdflib.term.Literal(
            "2000-11-02T12:27:00",
            datatype=rdflib.term.URIRef("http://www.w3.org/2001/XMLSchema#dateTime"),
        )
    )
    expected.append(
        rdflib.term.Literal(
            "2000-11-02",
            datatype=rdflib.term.URIRef("http://www.w3.org/2001/XMLSchema#date"),
        )
    )

    eq_(list(result2)[0][0], expected[0])
    eq_(list(result2)[1][0], expected[1])


def test_datetime_datetime_subs():
    """
    Test cases for subtraction operation between dateTime and dateTime

    """
    data = """
    @prefix : <#>.
    @prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
    @prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#>.
    @prefix xsd: <http://www.w3.org/2001/XMLSchema#>.
    """

    graph = Graph()
    f = io.StringIO(data)
    graph.parse(f, format="n3")

    result1 = graph.query(
        """
    PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
    SELECT (?l - ?r AS ?duration)
    WHERE {
        VALUES (?l ?r) {
            ("2000-10-30T06:12:00-05:00"^^xsd:dateTime "1999-11-28T09:00:00Z"^^xsd:dateTime)
            ("2000-10-30"^^xsd:date"1999-11-28"^^xsd:date)
        }
    }
    """
    )

    expected1 = rdflib.term.Literal(
        "P337DT2H12M",
        datatype=rdflib.term.URIRef("http://www.w3.org/2001/XMLSchema#duration"),
    )
    expected2 = rdflib.term.Literal(
        "P337D",
        datatype=rdflib.term.URIRef("http://www.w3.org/2001/XMLSchema#duration"),
    )

    eq_(list(result1)[0][0], expected1)
    eq_(list(result1)[1][0], expected2)



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class MF(DefinedNamespace):
    _fail = True
    _NS = Namespace("http://www.w3.org/2001/sw/DataAccess/tests/test-manifest#")

    IllFormedLiterals: (
        URIRef  # Tests that involve lexical forms which are illegal for the datatype
    )
    KnownTypesDefault2Neq: URIRef  # Values in disjoint value spaces are not equal
    LangTagAwareness: URIRef  # Tests that require langauge tag handling in FILTERs
    LaxCardinality: URIRef  # The given mf:result for a test with an mf:resultCardinality of mf:ReducedCardinalityTest  is the results as if the REDUCED keyword were omitted. To pass such a test, an implementation must produce a result set  with each solution in the expected results appearing at least once and  no more than the number of times it appears in the expected results. Of  course, there must also be no results produced that are not in the  expected results.
    Manifest: URIRef  # The class of manifests
    ManifestEntry: URIRef  # One entry in rdf:type list of entries
    NegativeSyntaxTest: URIRef  # A type of test specifically for syntax testing. Syntax tests are not required to have an associated result, only an action. Negative syntax tests are tests of which the result should be a parser error.
    NegativeSyntaxTest11: URIRef  # A type of test specifically for syntax testing of new features in the SPARQL1.1 Query Language. Syntax tests are not required to have an associated result, only an action. Negative syntax tests are tests of which the result should be a parser error.
    NegativeUpdateSyntaxTest11: URIRef  # A type of test specifically for syntax testing of SPARQL1.1 Update. Syntax tests are not required to have an associated result, only an action. Negative syntax tests are tests of which the result should be a parser error.
    Notable: URIRef  # Requirements for a  particular test
    PositiveSyntaxTest: URIRef  # A type of test specifically for syntax testing. Syntax tests are not required to have an associated result, only an action.
    PositiveSyntaxTest11: URIRef  # A type of test specifically for syntax testing of new features in the SPARQL1.1 Query Language. Syntax tests are not required to have an associated result, only an action.
    PositiveUpdateSyntaxTest11: URIRef  # A type of test specifically for syntax testing of SPARQL1.1 Update. Syntax tests are not required to have an associated result, only an action.
    QueryEvaluationTest: URIRef  # A type of test specifically for query evaluation testing. Query evaluation tests are required to have an associated input dataset, a query, and an expected output dataset.
    Requirement: URIRef  # Requirements for a  particular test
    ResultCardinality: URIRef  # Potential modes of evaluating a test's results with respect to solution cardinality
    StringSimpleLiteralCmp: URIRef  # Tests that require simple literal is the same value as an xsd:string of the same lexicial form
    TestStatus: URIRef  # Statuses a test can have
    UpdateEvaluationTest: URIRef  # The class of all SPARQL 1.1 Update evaluation tests
    XsdDateOperations: URIRef  # Tests that require xsd:date operations
    accepted: URIRef  # None
    action: URIRef  # Action to perform
    entries: URIRef  # Connects the manifest resource to rdf:type list of entries
    include: URIRef  # Connects the manifest resource to rdf:type list of manifests
    name: URIRef  # Optional name of this entry
    notable: URIRef  # Notable feature of this test (advisory)
    proposed: URIRef  # None
    rejected: URIRef  # None
    requires: URIRef  # Required functionality for execution of this test
    result: URIRef  # The expected outcome
    # result: URIRef  # The test status
    resultCardinality: URIRef  # Specifies whether passing the test requires strict or lax cardinality adherence
    CSVResultFormatTest: URIRef
    ServiceDescriptionTest: URIRef
    ProtocolTest: URIRef



from rdflib import FOAF, Graph, URIRef
from rdflib.compare import isomorphic


def test_skolem_de_skolem_roundtrip():
    """Test round-trip of skolemization/de-skolemization of data.

    Issue: https://github.com/RDFLib/rdflib/issues/1404
    """

    ttl = """
    @prefix wd: <http://www.wikidata.org/entity/> .
    @prefix foaf: <http://xmlns.com/foaf/0.1/> .

    wd:Q1203 foaf:knows [ a foaf:Person;
        foaf:name "Ringo" ].
    """

    graph = Graph()
    graph.parse(data=ttl, format="turtle")

    query = {
        "subject": URIRef("http://www.wikidata.org/entity/Q1203"),
        "predicate": FOAF.knows,
    }

    # Save the original bnode id.
    bnode_id = graph.value(**query)

    skolemized_graph = graph.skolemize()

    # Check the BNode is now a URIRef after skolemization.
    skolem_bnode = skolemized_graph.value(**query)
    assert type(skolem_bnode) is URIRef

    # Check that the original bnode id exists somewhere in the uri.
    assert bnode_id in skolem_bnode

    # Check that the original data is not isomorphic with the skolemized data.
    assert not isomorphic(graph, skolemized_graph)

    # Check that the original graph data is the same as the de-skolemized data.
    de_skolemized_graph = skolemized_graph.de_skolemize()
    assert isomorphic(graph, de_skolemized_graph)



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class DOAP(DefinedNamespace):
    """
    Description of a Project (DOAP) vocabulary

    The Description of a Project (DOAP) vocabulary, described using W3C RDF Schema and the Web Ontology Language.

    Generated from: http://usefulinc.com/ns/doap
    Date: 2024-08-01 13:03:22.175167
    """

    _NS = Namespace("http://usefulinc.com/ns/doap#")

    _fail = True

    ArchRepository: URIRef  # GNU Arch source code repository. Dépôt GNU Arch du code source. Repositorio GNU Arch del código fuente. GNU Arch Quellcode-Versionierungssystem. Úložiště zdrojových kódů GNU Arch. Repositório GNU Arch do código fonte.
    BKRepository: URIRef  # BitKeeper source code repository. Dépôt BitKeeper du code source. Repositorio BitKeeper del código fuente. BitKeeper Quellcode-Versionierungssystem. Úložiště zdrojových kódů BitKeeper. Repositório BitKeeper do código fonte.
    BazaarBranch: (
        URIRef  # Bazaar source code branch. Código fonte da ramificação Bazaar.
    )
    CVSRepository: URIRef  # CVS source code repository. Dépôt CVS du code source. Repositorio CVS del código fuente. CVS Quellcode-Versionierungssystem. Úložiště zdrojových kódů CVS. Repositório CVS do código fonte.
    DarcsRepository: URIRef  # darcs source code repository. Dépôt darcs du code source. Repositorio darcs del código fuente. Repositório darcs do código fonte.
    GitBranch: URIRef  # Git source code branch. Código fonte da ramificação Git.
    GitRepository: URIRef  # Git source code repository. Dépôt Git du code source. Repositorio Git del código fuente. Git Quellcode-Versionierungssystem. Úložiště zdrojových kódů Git. Repositório Git do código fonte.
    HgRepository: URIRef  # Mercurial source code repository. Repositório Mercurial do código fonte.
    Project: URIRef  # A project. Un projet. Un proyecto. Ein Projekt. Projekt. Projeto.
    Repository: URIRef  # Source code repository. Dépôt du code source. Repositorio del código fuente. Quellcode-Versionierungssystem. Úložiště zdrojových kódů. Repositório do código fonte.
    SVNRepository: URIRef  # Subversion source code repository. Dépôt Subversion du code source. Repositorio Subversion del código fuente. Subversion Quellcode-Versionierungssystem. Úložiště zdrojových kódů Subversion. Repositório Subversion do código fonte.
    Specification: URIRef  # A specification of a system's aspects, technical or otherwise. A especificação de aspetos, técnicas ou outros do sistema.
    Version: URIRef  # Version information of a project release. Détails sur une version d'une release d'un projet. Información sobre la versión de un release del proyecto. Versionsinformation eines Projekt Releases. Informace o uvolněné verzi projektu. Informação sobre a versão do projeto lançado.
    audience: (
        URIRef  # Description of target user base Descrição do utilizador base alvo
    )
    blog: URIRef  # URI of a blog related to a project URI de um blog relacionado com um projeto
    browse: URIRef  # Web browser interface to repository. Interface web au dépôt. Interface web del repositorio. Web-Browser Interface für das Repository. Webové rozhraní pro prohlížení úložiště. Interface web do repositório.
    category: URIRef  # A category of project. Une catégorie de projet. Una categoría de proyecto. Eine Kategorie eines Projektes. Kategorie projektu. Uma categoría de projeto.
    created: URIRef  # Date when something was created, in YYYY-MM-DD form. e.g. 2004-04-05 Date à laquelle a été créé quelque chose, au format AAAA-MM-JJ (par ex. 2004-04-05) Fecha en la que algo fue creado, en formato AAAA-MM-DD. e.g. 2004-04-05 Erstellungsdatum von Irgendwas, angegeben im YYYY-MM-DD Format, z.B. 2004-04-05. Datum, kdy bylo něco vytvořeno ve formátu RRRR-MM-DD, např. 2004-04-05 Data em que algo foi criado, no formato AAAA-MM-DD. e.g. 2004-04-05
    description: URIRef  # Plain text description of a project, of 2-4 sentences in length. Texte descriptif d'un projet, long de 2 à 4 phrases. Descripción en texto plano de un proyecto, de 2 a 4 enunciados de longitud. Beschreibung eines Projekts als einfacher Text mit der Länge von 2 bis 4 Sätzen. Čistě textový, 2 až 4 věty dlouhý popis projektu. Descrição de um projeto em texto apenas, com 2 a 4 frases de comprimento.
    developer: URIRef  # Developer of software for the project. Développeur pour le projet. Desarrollador de software para el proyecto. Software-Entwickler für das Projekt. Vývojář softwaru projektu. Programador de software para o projeto.
    documentation: (
        URIRef  # Documentation of the project. Aide pour l’utilisation de ce projet.
    )
    documenter: URIRef  # Contributor of documentation to the project. Collaborateur à la documentation du projet. Proveedor de documentación para el proyecto. Mitarbeiter an der Dokumentation des Projektes. Spoluautor dokumentace projektu. Contribuidor para a documentação do projeto.
    helper: URIRef  # Project contributor. Collaborateur au projet. Colaborador del proyecto. Projekt-Mitarbeiter. Spoluautor projektu. Ajudante ou colaborador do projeto.
    homepage: URIRef  # URL of a project's homepage, 		associated with exactly one project. L'URL de la page web d'un projet, 		associée avec un unique projet. El URL de la página de un proyecto, 		asociada con exactamente un proyecto. URL der Projekt-Homepage, 		verbunden mit genau einem Projekt. URL adresa domovské stránky projektu asociované s právě jedním projektem. O URL da página de um projeto, 		asociada com exactamente um projeto.
    implements: URIRef  # A specification that a project implements. Could be a standard, API or legally defined level of conformance. Uma especificação que um projeto implementa. Pode ser uma padrão, API ou um nível de conformidade definida legalmente.
    language: URIRef  # BCP47 language code a project has been translated into Código de idioma BCP47 do projeto para o qual foi traduzido
    license: URIRef  # The URI of an RDF description of the license the software is distributed under. E.g. a SPDX reference L'URI d'une description RDF de la licence sous laquelle le programme est distribué. El URI de una descripción RDF de la licencia bajo la cuál se distribuye el software. Die URI einer RDF-Beschreibung einer Lizenz unter der die Software herausgegeben wird. z.B. eine SPDX Referenz URI adresa RDF popisu licence, pod kterou je software distribuován. O URI de uma descrição RDF da licença do software sob a qual é distribuída. Ex.: referência SPDX
    location: URIRef  # Location of a repository. Emplacement d'un dépôt. lugar de un repositorio. Lokation eines Repositorys. Umístění úložiště. Localização de um repositório.
    maintainer: URIRef  # Maintainer of a project, a project leader. Développeur principal d'un projet, un meneur du projet. Desarrollador principal de un proyecto, un líder de proyecto. Hauptentwickler eines Projektes, der Projektleiter Správce projektu, vedoucí projektu. Programador principal de um projeto, um líder de projeto.
    module: URIRef  # Module name of a Subversion, CVS, BitKeeper or Arch repository. Nom du module d'un dépôt Subversion, CVS, BitKeeper ou Arch. Nombre del módulo de un repositorio Subversion, CVS, BitKeeper o Arch. Modul-Name eines Subversion, CVS, BitKeeper oder Arch Repositorys. Jméno modulu v CVS, BitKeeper nebo Arch úložišti. Nome do módulo de um repositório Subversion, CVS, BitKeeper ou Arch.
    name: URIRef  # A name of something. Le nom de quelque chose. El nombre de algo. Der Name von Irgendwas Jméno něčeho. O nome de alguma coisa.
    os: URIRef  # Operating system that a project is limited to.  Omit this property if the project is not OS-specific. Système d'exploitation auquel est limité le projet. Omettez cette propriété si le 		projet n'est pas limité à un système d'exploitation. Sistema opertivo al cuál está limitado el proyecto.  Omita esta propiedad si el proyecto no es específico 		de un sistema opertaivo en particular. Betriebssystem auf dem das Projekt eingesetzt werden kann. Diese Eigenschaft kann ausgelassen werden, wenn das Projekt nicht BS-spezifisch ist. Operační systém, na jehož použití je projekt limitován. Vynechejte tuto vlastnost, pokud je projekt nezávislý na operačním systému. Sistema operativo a que o projeto está limitado. Omita esta propriedade se o projeto não é condicionado pelo SO usado.
    platform: URIRef  # Indicator of software platform (non-OS specific), e.g. Java, Firefox, ECMA CLR Indicador da plataforma do software (não específico a nenhum SO), ex.: Java, Firefox, ECMA CLR
    release: URIRef  # A project release. Une release (révision) d'un projet. Un release (versión) de un proyecto. Ein Release (Version) eines Projekts. Relase (verze) projektu. A publicação de um projeto.
    repository: URIRef  # Source code repository. Dépôt du code source. Repositorio del código fuente. Quellcode-Versionierungssystem. Úložiště zdrojových kódů. Repositório do código fonte.
    repositoryOf: URIRef  # The project that uses a repository.
    revision: URIRef  # Revision identifier of a software release. Identifiant de révision d'une release du programme. Indentificador de la versión de un release de software. Versionsidentifikator eines Software-Releases. Identifikátor zpřístupněné revize softwaru. Identificador do lançamento da revisão do software.
    screenshots: URIRef  # Web page with screenshots of project. Page web avec des captures d'écran du projet. Página web con capturas de pantalla del proyecto. Web-Seite mit Screenshots eines Projektes. Webová stránka projektu se snímky obrazovky. Página web com as capturas de ecrãn do projeto.
    shortdesc: URIRef  # Short (8 or 9 words) plain text description of a project. Texte descriptif concis (8 ou 9 mots) d'un projet. Descripción corta (8 o 9 palabras) en texto plano de un proyecto. Kurzbeschreibung (8 oder 9 Wörter) eines Projekts als einfacher Text. Krátký (8 nebo 9 slov) čistě textový popis projektu. Descrição curta (com 8 ou 9 palavras) de um projeto em texto apenas.
    tester: URIRef  # A tester or other quality control contributor. Un testeur ou un collaborateur au contrôle qualité. Un tester u otro proveedor de control de calidad. Ein Tester oder anderer Mitarbeiter der Qualitätskontrolle. Tester nebo jiný spoluautor kontrolující kvalitu. Um controlador ou outro contribuidor para o controlo de qualidade.
    translator: URIRef  # Contributor of translations to the project. Collaborateur à la traduction du projet. Proveedor de traducciones al proyecto. Mitarbeiter an den Übersetzungen des Projektes. Spoluautor překladu projektu. Contribuidor das traduções para o projeto.
    vendor: URIRef  # Vendor organization: commercial, free or otherwise
    wiki: URIRef  # URL of Wiki for collaborative discussion of project. L'URL du Wiki pour la discussion collaborative sur le projet. URL del Wiki para discusión colaborativa del proyecto. Wiki-URL für die kollaborative Dikussion eines Projektes. URL adresa wiki projektu pro společné diskuse. URL da Wiki para discussão em grupo do projeto.

    # Valid non-python identifiers
    _extras = [
        "anon-root",  # Repository for anonymous access. Dépôt pour accès anonyme. Repositorio para acceso anónimo. Repository für anonymen Zugriff Úložiště pro anonymní přístup. Repositório para acesso anónimo.
        "bug-database",  # Bug tracker for a project. Suivi des bugs pour un projet. Bug tracker para un proyecto. Fehlerdatenbank eines Projektes. Správa chyb projektu. Bug tracker para um projeto.
        "developer-forum",  # A forum or community for developers of this project.
        "download-mirror",  # Mirror of software download web page. Miroir de la page de téléchargement du programme. Mirror de la página web de descarga. Spiegel der Seite von die Projekt-Software heruntergeladen werden kann. Zrcadlo stránky pro stažení softwaru. Mirror da página web para fazer download.
        "download-page",  # Web page from which the project software can be downloaded. Page web à partir de laquelle on peut télécharger le programme. Página web de la cuál se puede bajar el software. Web-Seite von der die Projekt-Software heruntergeladen werden kann. Webová stránka, na které lze stáhnout projektový software. Página web da qual o projeto de software pode ser descarregado.
        "file-release",  # URI of download associated with this release. URI adresa stažení asociované s revizí. URI para download associado com a publicação.
        "mailing-list",  # Mailing list home page or email address. Page web de la liste de diffusion, ou adresse de courriel. Página web de la lista de correo o dirección de correo. Homepage der Mailing Liste oder E-Mail Adresse. Domovská stránka nebo e–mailová adresa e–mailové diskuse. Página web da lista de distribuição de e-mail ou dos endereços.
        "old-homepage",  # URL of a project's past homepage, 		associated with exactly one project. L'URL d'une ancienne page web d'un 		projet, associée avec un unique projet. El URL de la antigua página de un proyecto, 		asociada con exactamente un proyecto. URL der letzten Projekt-Homepage, 		verbunden mit genau einem Projekt. URL adresa předešlé domovské stránky projektu asociované s právě jedním projektem. O URL antigo da página de um projeto, 		associada com exactamente um projeto.
        "programming-language",  # Programming language a project is implemented in or intended for use with. Langage de programmation avec lequel un projet est implémenté, 		ou avec lequel il est prévu de l'utiliser. Lenguaje de programación en el que un proyecto es implementado o con el cuál pretende usarse. Programmiersprache in der ein Projekt implementiert ist oder intendiert wird zu benutzen. Programovací jazyk, ve kterém je projekt implementován nebo pro který je zamýšlen k použití. Linguagem de programação que o projeto usa ou é para ser utilizada.
        "security-contact",  # The Agent that should be contacted 	if security issues are found with the project.
        "security-policy",  # URL of the security policy of a project.
        "service-endpoint",  # The URI of a web service endpoint where software as a service may be accessed
        "support-forum",  # A forum or community that supports this project.
    ]



import base64

from rdflib import XSD, Literal


class TestB64Binary:
    def test_unicode(self):
        str1 = "Test utf-8 string éàë"
        # u b64string
        b64_str1 = base64.b64encode(str1.encode("utf-8")).decode()
        l1 = Literal(b64_str1, datatype=XSD.base64Binary)
        b_str1 = l1.toPython()
        assert b_str1.decode("utf-8") == str1
        assert str(l1) == b64_str1

        # b b64string
        b64_str1b = base64.b64encode(str1.encode("utf-8"))
        l1b = Literal(b64_str1b, datatype=XSD.base64Binary)
        b_str1b = l1b.toPython()
        assert b_str1 == b_str1b
        assert b_str1b.decode("utf-8") == str1
        assert str(l1b) == b64_str1



"""
This module is intended for tests related to unquoting/unescaping in various
formats that are related to turtle, such as ntriples, nquads, trig and n3.
"""

from __future__ import annotations

import itertools
import logging
import re
from typing import Callable, Dict, Iterable, List, Tuple

import pytest

from rdflib.graph import ConjunctiveGraph, Graph
from rdflib.plugins.parsers import ntriples
from rdflib.term import Literal, URIRef
from test.utils.namespace import EGDC

from .utils import GraphHelper

# https://www.w3.org/TR/turtle/#string
string_escape_map = {
    "t": "\t",
    "b": "\b",
    "n": "\n",
    "r": "\r",
    "f": "\f",
    '"': '"',
    "'": "'",
    "\\": "\\",
}


def make_unquote_correctness_pairs() -> List[Tuple[str, str]]:
    """
    Creates pairs of quoted and unquoted strings.
    """
    result = []

    def add_pair(escape: str, unescaped: str) -> None:
        result.append((f"\\{escape}", unescaped))
        result.append((f"\\\\{escape}", f"\\{escape}"))
        result.append((f"\\\\\\{escape}", f"\\{unescaped}"))

    chars = "A1a\\\nøæå"
    for char in chars:
        code_point = ord(char)
        add_pair(f"u{code_point:04x}", char)
        add_pair(f"u{code_point:04X}", char)
        add_pair(f"U{code_point:08x}", char)
        add_pair(f"U{code_point:08X}", char)

    string_escapes = "tbnrf'"
    for char in string_escapes:
        add_pair(f"{char}", string_escape_map[char])

    # special handling because «"» should not appear in string, and add_pair
    # will add it.
    result.append(('\\"', '"'))
    result.append(('\\\\\\"', '\\"'))

    # special handling because «\» should not appear in string, and add_pair
    # will add it.
    result.append(("\\\\", "\\"))
    result.append(("\\\\\\\\", "\\\\"))

    return result


UNQUOTE_CORRECTNESS_PAIRS = make_unquote_correctness_pairs()


def ntriples_unquote_validate(input: str) -> str:
    """
    This function wraps `ntriples.unquote` in a way that ensures that `ntriples.validate` is always ``True`` when it runs.
    """
    old_validate = ntriples.validate
    try:
        ntriples.validate = True
        return ntriples.unquote(input)
    finally:
        ntriples.validate = old_validate


def ntriples_unquote(input: str) -> str:
    """
    This function wraps `ntriples.unquote` in a way that ensures that `ntriples.validate` is always ``False`` when it runs.
    """
    old_validate = ntriples.validate
    try:
        ntriples.validate = False
        return ntriples.unquote(input)
    finally:
        ntriples.validate = old_validate


unquoters: Dict[str, Callable[[str], str]] = {
    "ntriples_unquote": ntriples_unquote,
    "ntriples_unquote_validate": ntriples_unquote_validate,
}


def make_unquote_correctness_tests(
    selectors: Iterable[str],
) -> Iterable[Tuple[str, str, str]]:
    """
    This function creates a cartesian product of the selectors and
    `CORRECTNESS_PAIRS` that is suitable for use as pytest parameters.
    """
    for selector in selectors:
        for quoted, unquoted in UNQUOTE_CORRECTNESS_PAIRS:
            yield selector, quoted, unquoted


@pytest.mark.parametrize(
    "unquoter_key, quoted, unquoted", make_unquote_correctness_tests(unquoters.keys())
)
def test_unquote_correctness(
    unquoter_key: str,
    quoted: str,
    unquoted: str,
) -> None:
    """
    Various unquote functions work correctly.
    """
    unquoter = unquoters[unquoter_key]
    assert unquoted == unquoter(quoted)


QUAD_FORMATS = {"nquads"}


@pytest.mark.parametrize(
    "format, quoted, unquoted",
    make_unquote_correctness_tests(["turtle", "ntriples", "nquads"]),
)
def test_parse_correctness(
    format: str,
    quoted: str,
    unquoted: str,
) -> None:
    """
    Quoted strings parse correctly
    """
    if format in QUAD_FORMATS:
        data = f'<example:Subject> <example:Predicate> "{quoted}" <example:Graph>.'
    else:
        data = f'<example:Subject> <example:Predicate> "{quoted}".'
    graph = ConjunctiveGraph()
    graph.parse(data=data, format=format)
    objs = list(graph.objects())
    assert len(objs) == 1
    obj = objs[0]
    assert isinstance(obj, Literal)
    assert isinstance(obj.value, str)
    assert obj.value == unquoted


@pytest.mark.parametrize(
    "format, char, escaped",
    [
        (format, char, escaped)
        for format, (char, escaped) in itertools.product(
            ["turtle"],
            [
                (r"x", r"x"),
                (r"(", r"\("),
                (r")", r"\)"),
            ],
        )
    ],
)
def test_pname_escaping(format: str, char: str, escaped: str) -> None:
    graph = Graph()
    triple = (
        URIRef(EGDC["prefix/John_Doe"]),
        URIRef(EGDC[f"prefix/prop{char}"]),
        Literal("foo", lang="en"),
    )
    graph.bind("egns", EGDC["prefix/"])
    graph.add(triple)
    data = graph.serialize(format=format)
    pattern = re.compile(f"\\segns:prop{re.escape(escaped)}\\s")
    logging.debug(
        "format = %s, char = %r, escaped = %r, pattern = %r, data = %s",
        format,
        char,
        escaped,
        pattern,
        data,
    )
    assert re.search(pattern, data) is not None


# https://www.w3.org/TR/turtle/#grammar-production-PN_LOCAL_ESC
# Not including %, as % should be used for percent encoding.
PN_LOCAL_ESC_CHARS = r"_~.-!$&'()*+,;=/?#@"


@pytest.mark.parametrize(
    "format, char",
    itertools.product(["turtle", "ntriples"], "A2c" + PN_LOCAL_ESC_CHARS),
)
def test_serialize_roundtrip(format: str, char: str) -> None:
    graph = Graph()
    triple = (
        URIRef(EGDC["prefix/John_Doe"]),
        URIRef(EGDC[f"prefix/prop{char}"]),
        Literal("foo", lang="en"),
    )
    graph.add(triple)
    graph.bind("egns", EGDC["prefix/"])
    data = graph.serialize(format=format)
    logging.debug("format = %s, char = %s, data = %s", format, char, data)
    parsed_graph = Graph()
    parsed_graph.parse(data=data, format=format)
    GraphHelper.assert_sets_equals(graph, parsed_graph)



from pathlib import Path

from rdflib import BNode, ConjunctiveGraph, Dataset, Literal, URIRef
from rdflib.compare import isomorphic
from rdflib.graph import DATASET_DEFAULT_GRAPH_ID
from rdflib.namespace import XSD


def test_named_and_anonymous_graph_roundtrip():
    s = """
        ["http://example.com/s01", "http://example.com/a", "http://example.com/Type1", "globalId", "", "https://example.com/graph/1"]
        ["http://example.com/s01", "http://example.com/label", "This is a Label", "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString", "en", "_:graph-2"]
        ["http://example.com/s01", "http://example.com/comment", "This is a comment", "http://www.w3.org/2001/XMLSchema#string", "", ""]
    """
    d = Dataset()
    d.parse(data=s, format="hext")

    new_s = d.serialize(format="hext")
    new_d = Dataset()
    new_d.parse(data=new_s, format="hext")

    named_graph = URIRef("https://example.com/graph/1")
    assert isomorphic(d.graph(named_graph), new_d.graph(named_graph))

    anonymous_graph = BNode("graph-2")
    assert isomorphic(d.graph(anonymous_graph), new_d.graph(anonymous_graph))

    assert isomorphic(
        d.graph(DATASET_DEFAULT_GRAPH_ID), new_d.graph(DATASET_DEFAULT_GRAPH_ID)
    )


def test_small_string():
    s = """
        ["http://example.com/s01", "http://example.com/a", "http://example.com/Type1", "globalId", "", "https://example.com/graph/1"]
        ["http://example.com/s01", "http://example.com/label", "This is a Label", "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString", "en", "_:graph-2"]
        ["http://example.com/s01", "http://example.com/comment", "This is a comment", "http://www.w3.org/2001/XMLSchema#string", "", ""]
        ["http://example.com/s01", "http://example.com/creationDate", "2021-12-01", "http://www.w3.org/2001/XMLSchema#date", "", ""]
        ["http://example.com/s01", "http://example.com/creationTime", "2021-12-01T12:13:00", "http://www.w3.org/2001/XMLSchema#dateTime", "", ""]
        ["http://example.com/s01", "http://example.com/age", "42", "http://www.w3.org/2001/XMLSchema#integer", "", ""]
        ["http://example.com/s01", "http://example.com/trueFalse", "false", ",http://www.w3.org/2001/XMLSchema#boolean", "", ""]
        ["http://example.com/s01", "http://example.com/op1", "http://example.com/o1", "globalId", "", ""]
        ["http://example.com/s01", "http://example.com/op1", "http://example.com/o2", "globalId", "", ""]
        ["http://example.com/s01", "http://example.com/op2", "http://example.com/o3", "globalId", "", ""]
        """
    d = Dataset()
    d.parse(data=s, format="hext")

    expected_graph_names = (
        URIRef(DATASET_DEFAULT_GRAPH_ID),
        URIRef("https://example.com/graph/1"),
        BNode("graph-2"),
    )
    for graph in d.contexts():
        assert graph.identifier in expected_graph_names

    assert len(d) == 10


def test_small_bytes_string():
    s = b"""\
        ["http://example.com/s01", "http://example.com/a", "http://example.com/Type1", "globalId", "", "https://example.com/graph/1"]
        ["http://example.com/s01", "http://example.com/label", "This is a Label", "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString", "en", "_:graph-2"]
        ["http://example.com/s01", "http://example.com/comment", "This is a comment", "http://www.w3.org/2001/XMLSchema#string", "", ""]
        ["http://example.com/s01", "http://example.com/creationDate", "2021-12-01", "http://www.w3.org/2001/XMLSchema#date", "", ""]
        ["http://example.com/s01", "http://example.com/creationTime", "2021-12-01T12:13:00", "http://www.w3.org/2001/XMLSchema#dateTime", "", ""]
        ["http://example.com/s01", "http://example.com/age", "42", "http://www.w3.org/2001/XMLSchema#integer", "", ""]
        ["http://example.com/s01", "http://example.com/trueFalse", "false", ",http://www.w3.org/2001/XMLSchema#boolean", "", ""]
        ["http://example.com/s01", "http://example.com/op1", "http://example.com/o1", "globalId", "", ""]
        ["http://example.com/s01", "http://example.com/op1", "http://example.com/o2", "globalId", "", ""]
        ["http://example.com/s01", "http://example.com/op2", "http://example.com/o3", "globalId", "", ""]
        """
    d = Dataset()
    d.parse(data=s, format="hext")

    expected_graph_names = (
        URIRef(DATASET_DEFAULT_GRAPH_ID),
        URIRef("https://example.com/graph/1"),
        BNode("graph-2"),
    )
    for graph in d.contexts():
        assert graph.identifier in expected_graph_names

    assert len(d) == 10


def test_small_string_cg():
    s = """
        ["http://example.com/s01", "http://example.com/a", "http://example.com/Type1", "globalId", "", "https://example.com/graph/1"]
        ["http://example.com/s01", "http://example.com/label", "This is a Label", "http://www.w3.org/1999/02/22-rdf-syntax-ns#langString", "en", "_:graph-2"]
        ["http://example.com/s01", "http://example.com/comment", "This is a comment", "http://www.w3.org/2001/XMLSchema#string", "", ""]
        ["http://example.com/s01", "http://example.com/creationDate", "2021-12-01", "http://www.w3.org/2001/XMLSchema#date", "", ""]
        ["http://example.com/s01", "http://example.com/creationTime", "2021-12-01T12:13:00", "http://www.w3.org/2001/XMLSchema#dateTime", "", ""]
        ["http://example.com/s01", "http://example.com/age", "42", "http://www.w3.org/2001/XMLSchema#integer", "", ""]
        ["http://example.com/s01", "http://example.com/trueFalse", "false", ",http://www.w3.org/2001/XMLSchema#boolean", "", ""]
        ["http://example.com/s01", "http://example.com/op1", "http://example.com/o1", "globalId", "", ""]
        ["http://example.com/s01", "http://example.com/op1", "http://example.com/o2", "globalId", "", ""]
        ["http://example.com/s01", "http://example.com/op2", "http://example.com/o3", "globalId", "", ""]
        """
    d = ConjunctiveGraph(identifier=DATASET_DEFAULT_GRAPH_ID)
    d.parse(data=s, format="hext")

    expected_graph_names = (
        URIRef(DATASET_DEFAULT_GRAPH_ID),
        URIRef("https://example.com/graph/1"),
        BNode("graph-2"),
    )
    for graph in d.contexts():
        assert graph.identifier in expected_graph_names

    assert len(d) == 10


def test_small_file_singlegraph():
    d = Dataset().parse(
        Path(__file__).parent.parent / "data/test_parser_hext_singlegraph.ndjson",
        format="hext",
    )
    assert len(d) == 10


def test_small_file_multigraph():
    d = Dataset()
    assert len(d) == 0
    d.parse(
        Path(__file__).parent.parent / "data/test_parser_hext_multigraph.ndjson",
        format="hext",
        publicID=d.default_context.identifier,
    )

    """There are 22 lines in the file test_parser_hext_multigraph.ndjson. When loaded
    into a Dataset, we get only 18 quads since the the dataset can contextualise
    the triples and thus deduplicate 4."""
    total_triples = 0
    # count all the triples in the Dataset
    for context in d.contexts():
        for triple in context.triples((None, None, None)):
            total_triples += 1
    assert total_triples == 18


def test_small_file_multigraph_cg():
    d = ConjunctiveGraph()
    assert len(d) == 0
    d.parse(
        Path(__file__).parent.parent / "data/test_parser_hext_multigraph.ndjson",
        format="hext",
        publicID=d.default_context.identifier,
    )

    """There are 22 lines in the file test_parser_hext_multigraph.ndjson. When loaded
    into a CG, we get only 18 quads since the the CG can contextualise
    the triples and thus deduplicate 4."""
    total_triples = 0
    # count all the triples in the Dataset
    for context in d.contexts():
        for triple in context.triples((None, None, None)):
            total_triples += 1
    assert total_triples == 18


def test_roundtrip():
    # these are some RDF files that HexT can round-trip since the have no
    # literals with no datatype declared:
    TEST_DIR = Path(__file__).parent.absolute() / "nt"  # noqa: N806
    files_to_skip = {
        "paths-04.nt": "subject literal",
        "even_more_literals.nt": "JSON decoding error",
        "literals-02.nt": "JSON decoding error",
        "more_literals.nt": "JSON decoding error",
        "test.nt": "JSON decoding error",
        "literals-05.nt": "JSON decoding error",
        "i18n-01.nt": "JSON decoding error",
        "literals-04.nt": "JSON decoding error",
        "rdflibtest01.nt": "JSON decoding error",
        "rdflibtest05.nt": "JSON decoding error",
    }
    tests = 0
    skipped = 0
    skip = False
    print()
    p = TEST_DIR.glob("**/*")
    for f in [x for x in p if x.is_file()]:
        tests += 1
        print(f"Test {tests}: {f}")
        if f.name not in files_to_skip.keys():
            try:
                cg = ConjunctiveGraph().parse(f, format="nt")
                # print(cg.serialize(format="n3"))
            except Exception:
                print("Skipping: could not NT parse")
                skipped += 1
                skip = True
            if not skip:
                cg2 = ConjunctiveGraph()
                cg2.parse(
                    data=cg.serialize(format="hext"),
                    format="hext",
                    publicID=cg2.default_context.identifier,
                )
                if cg2.context_aware:
                    for context in cg2.contexts():
                        for triple in context.triples((None, None, None)):
                            if type(triple[2]) is Literal:
                                if triple[2].datatype == XSD.string:
                                    context.remove((triple[0], triple[1], triple[2]))
                                    context.add(
                                        (triple[0], triple[1], Literal(str(triple[2])))
                                    )
                else:
                    for triple in cg2.triples((None, None, None)):
                        if type(triple[2]) is Literal:
                            if triple[2].datatype == XSD.string:
                                cg2.remove((triple[0], triple[1], triple[2]))
                                cg2.add((triple[0], triple[1], Literal(str(triple[2]))))

                # print(cg2.serialize(format="trig"))
                assert cg.isomorphic(cg2)
            skip = False
        else:
            print(f"Skipping: {files_to_skip[f.name]}")

    print(f"No. tests: {tests}")
    print(f"No. tests skipped: {skipped}")


if __name__ == "__main__":
    test_small_file_multigraph()



import pytest

from rdflib.graph import Graph
from rdflib.namespace import RDFS
from rdflib.term import BNode, Literal

remove_me = (BNode(), RDFS.label, Literal("remove_me"))


@pytest.fixture(scope="function")
def get_store(request):
    store = Graph(store="default")
    store.open("store")
    store.add(remove_me)

    yield store

    store.close()


def test_add(get_store):
    store = get_store
    subject = BNode()
    store.add((subject, RDFS.label, Literal("foo")))


def test_remove(get_store):
    store = get_store
    store.remove(remove_me)
    store.remove((None, None, None))


def test_triples(get_store):
    store = get_store
    for s, p, o in store:
        pass



from rdflib import Dataset, Graph, Literal, URIRef


def test_add_quad():
    ds = Dataset()
    ds.add(
        (
            URIRef("http://example.org/subject1"),
            URIRef("http://example.org/predicate2"),
            Literal("object2"),
            Graph(identifier=URIRef("http://example.org/graph1")),
        )
    )
    result = ds.serialize(format="patch", operation="add")
    assert (
        """A <http://example.org/subject1> <http://example.org/predicate2> "object2" <http://example.org/graph1> .
"""
        in result
    )


def test_delete_quad():
    ds = Dataset()
    ds.add(
        (
            URIRef("http://example.org/subject1"),
            URIRef("http://example.org/predicate2"),
            Literal("object2"),
            Graph(identifier=URIRef("http://example.org/graph1")),
        )
    )
    result = ds.serialize(format="patch", operation="remove")
    assert (
        """D <http://example.org/subject1> <http://example.org/predicate2> "object2" <http://example.org/graph1> .
"""
        in result
    )


def test_diff_quad():
    quad_1 = (
        URIRef("http://example.org/subject1"),
        URIRef("http://example.org/predicate2"),
        Literal("object2"),
        Graph(identifier=URIRef("http://example.org/graph1")),
    )
    quad_2 = (
        URIRef("http://example.org/subject2"),
        URIRef("http://example.org/predicate3"),
        Literal("object3"),
        Graph(identifier=URIRef("http://example.org/graph2")),
    )
    ds1 = Dataset()
    ds2 = Dataset()
    ds1.add(quad_1)
    ds2.addN([quad_1, quad_2])
    result = ds1.serialize(format="patch", target=ds2)
    assert (
        """A <http://example.org/subject2> <http://example.org/predicate3> "object3" <http://example.org/graph2> ."""
        in result
    )


def test_add_triple():
    ds = Dataset()
    ds.add(
        (
            URIRef("http://example.org/subject1"),
            URIRef("http://example.org/predicate2"),
            Literal("object2"),
        )
    )
    result = ds.serialize(format="patch", operation="add")
    assert (
        """A <http://example.org/subject1> <http://example.org/predicate2> "object2" ."""
        in result
    )


def test_delete_triple():
    ds = Dataset()
    ds.add(
        (
            URIRef("http://example.org/subject1"),
            URIRef("http://example.org/predicate2"),
            Literal("object2"),
        )
    )
    result = ds.serialize(format="patch", operation="remove")
    assert (
        """D <http://example.org/subject1> <http://example.org/predicate2> "object2" ."""
        in result
    )


def test_diff_triple():
    triple_1 = (
        URIRef("http://example.org/subject1"),
        URIRef("http://example.org/predicate2"),
        Literal("object2"),
    )
    triple_2 = (
        URIRef("http://example.org/subject2"),
        URIRef("http://example.org/predicate3"),
        Literal("object3"),
    )
    ds1 = Dataset()
    ds2 = Dataset()
    ds1.add(triple_1)
    ds2.add(triple_1)
    ds2.add(triple_2)
    result = ds1.serialize(format="patch", target=ds2)
    assert (
        """A <http://example.org/subject2> <http://example.org/predicate3> "object3" ."""
        in result
    )


def test_diff_quad_overlap():
    quad_1 = (
        URIRef("http://example.org/subject1"),
        URIRef("http://example.org/predicate1"),
        Literal("object1"),
        Graph(identifier=URIRef("http://example.org/graph1")),
    )
    quad_2 = (
        URIRef("http://example.org/subject2"),
        URIRef("http://example.org/predicate2"),
        Literal("object2"),
        Graph(identifier=URIRef("http://example.org/graph2")),
    )
    quad_3 = (
        URIRef("http://example.org/subject3"),
        URIRef("http://example.org/predicate3"),
        Literal("object3"),
        Graph(identifier=URIRef("http://example.org/graph3")),
    )
    ds1 = Dataset()
    ds2 = Dataset()
    ds1.addN([quad_1, quad_2])
    ds2.addN([quad_2, quad_3])
    result = ds1.serialize(format="patch", target=ds2)
    # first quad needs to be removed
    assert (
        """D <http://example.org/subject1> <http://example.org/predicate1> "object1" <http://example.org/graph1> ."""
        in result
    )
    # third quad needs to be added
    assert (
        """A <http://example.org/subject3> <http://example.org/predicate3> "object3" <http://example.org/graph3> ."""
        in result
    )


def test_header_id():
    ds = Dataset()
    ds.add(
        (
            URIRef("http://example.org/subject1"),
            URIRef("http://example.org/predicate2"),
            Literal("object2"),
        )
    )
    result = ds.serialize(format="patch", operation="add", header_id="uuid:123")
    assert """H id <uuid:123>""" in result


def test_prev_header():
    ds = Dataset()
    ds.add(
        (
            URIRef("http://example.org/subject1"),
            URIRef("http://example.org/predicate2"),
            Literal("object2"),
        )
    )
    result = ds.serialize(format="patch", operation="add", header_prev="uuid:123")
    assert """H prev <uuid:123>""" in result



"""
some more specific Literal tests are in test_literal.py
"""

import base64
import random

from rdflib.graph import Graph, QuotedGraph
from rdflib.namespace import XSD
from rdflib.term import BNode, Literal, URIRef, _is_valid_unicode


def uformat(s):
    return s.replace("u'", "'")


class TestURIRefRepr:
    """
    see also test_literal.TestRepr
    """

    def test_subclass_name_appears_in_repr(self):
        class MyURIRef(URIRef):
            pass

        x = MyURIRef("http://example.com/")
        assert repr(x) == uformat("MyURIRef(u'http://example.com/')")

    def test_graceful_ordering(self):
        u = URIRef("cake")
        g = Graph()
        a = u > u
        a = u > BNode()
        a = u > QuotedGraph(g.store, u)
        a = u > g  # noqa: F841


class TestBNodeRepr:
    def test_subclass_name_appears_in_repr(self):
        class MyBNode(BNode):
            pass

        x = MyBNode()
        assert repr(x).startswith("MyBNode(")


class TestLiteral:
    def test_base64_values(self):
        b64msg = "cmRmbGliIGlzIGNvb2whIGFsc28gaGVyZSdzIHNvbWUgYmluYXJ5IAAR83UC"
        decoded_b64msg = base64.b64decode(b64msg)
        lit = Literal(b64msg, datatype=XSD.base64Binary)
        assert lit.value == decoded_b64msg
        assert str(lit) == b64msg

    def test_total_order(self):
        types = {
            XSD.dateTime: (
                "2001-01-01T00:00:00",
                "2001-01-01T00:00:00Z",
                "2001-01-01T00:00:00-00:00",
            ),
            XSD.date: ("2001-01-01", "2001-01-01Z", "2001-01-01-00:00"),
            XSD.time: ("00:00:00", "00:00:00Z", "00:00:00-00:00"),
            XSD.gYear: ("2001", "2001Z", "2001-00:00"),  # interval
            XSD.gYearMonth: ("2001-01", "2001-01Z", "2001-01-00:00"),
        }
        literals = [
            Literal(literal, datatype=t)
            for t, literals in types.items()
            for literal in literals
        ]
        try:
            sorted(literals)
            orderable = True
        except TypeError as e:
            for l_ in literals:
                print(repr(l_), repr(l_.value))
            print(e)
            orderable = False
        assert orderable is True

        # also make sure that within a datetime things are still ordered:
        l1 = [
            Literal(l_, datatype=XSD.dateTime)
            for l_ in [
                "2001-01-01T00:00:00",
                "2001-01-01T01:00:00",
                "2001-01-01T01:00:01",
                "2001-01-02T01:00:01",
                "2001-01-01T00:00:00Z",
                "2001-01-01T00:00:00-00:00",
                "2001-01-01T01:00:00Z",
                "2001-01-01T01:00:00-00:00",
                "2001-01-01T00:00:00-01:30",
                "2001-01-01T01:00:00-01:30",
                "2001-01-02T01:00:01Z",
                "2001-01-02T01:00:01-00:00",
                "2001-01-02T01:00:01-01:30",
            ]
        ]
        l2 = list(l1)
        random.shuffle(l2)
        assert l1 == sorted(l2)

    def test_literal_add(self):
        from decimal import Decimal

        # compares Python decimals
        def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
            a = float(a)
            b = float(b)
            return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)

        cases = [
            (1, Literal(1), Literal(1), Literal(2)),
            (2, Literal(Decimal(1)), Literal(Decimal(1)), Literal(Decimal(2))),
            (3, Literal(float(1)), Literal(float(1)), Literal(float(2))),
            (4, Literal(1), Literal(1.1), Literal(2.1, datatype=XSD.decimal)),
            (5, Literal(1.1), Literal(1.1), Literal(2.2)),
            (
                6,
                Literal(Decimal(1)),
                Literal(Decimal(1.1)),
                Literal(Decimal(2.1), datatype=XSD.decimal),
            ),
            (7, Literal(Decimal(1.1)), Literal(Decimal(1.1)), Literal(Decimal(2.2))),
            (8, Literal(float(1)), Literal(1.1), Literal(2.1)),
            (9, Literal(1.1), Literal(1.1), Literal(2.2)),
            (10, Literal(-1), Literal(-1), Literal(-2)),
            (12, Literal(Decimal(-1)), Literal(Decimal(-1)), Literal(Decimal(-2))),
            (13, Literal(float(-1)), Literal(float(-1)), Literal(float(-2))),
            (14, Literal(-1), Literal(-1.1), Literal(-2.1)),
            (15, Literal(-1.1), Literal(-1.1), Literal(-2.2)),
            (16, Literal(Decimal(-1)), Literal(Decimal(-1.1)), Literal(Decimal(-2.1))),
            (
                17,
                Literal(Decimal(-1.1)),
                Literal(Decimal(-1.1)),
                Literal(Decimal(-2.2)),
            ),
            (18, Literal(float(-1)), Literal(-1.1), Literal(-2.1)),
            (19, Literal(-1.1), Literal(-1.1), Literal(-2.2)),
            (20, Literal(1), Literal(1.0), Literal(2.0)),
            (21, Literal(1.0), Literal(1.0), Literal(2.0)),
            (22, Literal(Decimal(1)), Literal(Decimal(1.0)), Literal(Decimal(2.0))),
            (23, Literal(Decimal(1.0)), Literal(Decimal(1.0)), Literal(Decimal(2.0))),
            (24, Literal(float(1)), Literal(1.0), Literal(2.0)),
            (25, Literal(1.0), Literal(1.0), Literal(2.0)),
            (
                26,
                Literal(1, datatype=XSD.integer),
                Literal(1, datatype=XSD.integer),
                Literal(2, datatype=XSD.integer),
            ),
            (
                27,
                Literal(1, datatype=XSD.integer),
                Literal("1", datatype=XSD.integer),
                Literal("2", datatype=XSD.integer),
            ),
            (
                28,
                Literal("1", datatype=XSD.integer),
                Literal("1", datatype=XSD.integer),
                Literal("2", datatype=XSD.integer),
            ),
            (
                29,
                Literal("1"),
                Literal("1", datatype=XSD.integer),
                Literal("11", datatype=XSD.string),
            ),
            (
                30,
                Literal(1),
                Literal("1", datatype=XSD.integer),
                Literal("2", datatype=XSD.integer),
            ),
            (
                31,
                Literal(Decimal(1), datatype=XSD.decimal),
                Literal(Decimal(1), datatype=XSD.decimal),
                Literal(Decimal(2), datatype=XSD.decimal),
            ),
            (
                32,
                Literal(Decimal(1)),
                Literal(Decimal(1), datatype=XSD.decimal),
                Literal(Decimal(2), datatype=XSD.decimal),
            ),
            (
                33,
                Literal(float(1)),
                Literal(float(1), datatype=XSD.float),
                Literal(float(2), datatype=XSD.float),
            ),
            (
                34,
                Literal(float(1), datatype=XSD.float),
                Literal(float(1), datatype=XSD.float),
                Literal(float(2), datatype=XSD.float),
            ),
            (35, Literal(1), 1, Literal(2)),
            (36, Literal(1), 1.0, Literal(2, datatype=XSD.decimal)),
            (37, Literal(1.0), 1, Literal(2, datatype=XSD.decimal)),
            (38, Literal(1.0), 1.0, Literal(2.0)),
            (39, Literal(Decimal(1.0)), Decimal(1), Literal(Decimal(2.0))),
            (40, Literal(Decimal(1.0)), Decimal(1.0), Literal(Decimal(2.0))),
            (41, Literal(1.0), float(1), Literal(2.0)),
            (42, Literal(1.0), 1.0, Literal(2.0)),
            (
                43,
                Literal(1, datatype=XSD.integer),
                "+1.1",
                Literal("1+1.1", datatype=XSD.string),
            ),
            (
                44,
                Literal(1, datatype=XSD.integer),
                Literal("+1.1", datatype=XSD.string),
                Literal("1+1.1", datatype=XSD.string),
            ),
            (
                45,
                Literal(Decimal(1.0), datatype=XSD.integer),
                Literal("1", datatype=XSD.string),
                Literal("11", datatype=XSD.string),
            ),
            (
                46,
                Literal(1.1, datatype=XSD.integer),
                Literal("1", datatype=XSD.string),
                Literal("1.11", datatype=XSD.string),
            ),
            (
                47,
                Literal(1, datatype=XSD.integer),
                None,
                Literal(1, datatype=XSD.integer),
            ),
            (
                48,
                Literal("1", datatype=XSD.string),
                None,
                Literal("1", datatype=XSD.string),
            ),
        ]

        for case in cases:
            # see if the addition exactly matches the expected output
            case_passed = (case[1] + case[2]) == (case[3])
            # see if the addition almost matches the expected output, for decimal precision errors
            if not case_passed:
                try:
                    case_passed = isclose((case[1] + case[2].value), case[3].value)
                except Exception:
                    pass

            if not case_passed:
                print(case[1], case[2])
                print("expected: " + case[3] + ", " + case[3].datatype)
                print(
                    "actual: "
                    + (case[1] + case[2])
                    + ", "
                    + (case[1] + case[2]).datatype
                )

            assert case_passed, "Case " + str(case[0]) + " failed"


class TestValidityFunctions:
    def test_is_valid_unicode(self):
        testcase_list = (
            (None, True),
            (1, True),
            (["foo"], True),
            ({"foo": b"bar"}, True),
            ("foo", True),
            (b"foo\x00", True),
            (b"foo\xf3\x02", False),
        )
        for val, expected in testcase_list:
            assert _is_valid_unicode(val) == expected



"""
SPARQL implementation for RDFLib

.. versionadded:: 4.0
"""

from importlib.metadata import entry_points
from typing import TYPE_CHECKING

SPARQL_LOAD_GRAPHS = True
"""
If True, using FROM <uri> and FROM NAMED <uri>
will load/parse more data
"""


SPARQL_DEFAULT_GRAPH_UNION = True
"""
If True - the default graph in the RDF Dataset is the union of all
named graphs (like RDFLib's ConjunctiveGraph)
"""


CUSTOM_EVALS = {}
"""
Custom evaluation functions

These must be functions taking (ctx, part) and raise
NotImplementedError if they cannot handle a certain part
"""


PLUGIN_ENTRY_POINT = "rdf.plugins.sparqleval"


from . import operators, parser, parserutils
from .processor import prepareQuery, prepareUpdate, processUpdate

assert parser
assert operators
assert parserutils


all_entry_points = entry_points()
if hasattr(all_entry_points, "select"):
    for ep in all_entry_points.select(group=PLUGIN_ENTRY_POINT):
        CUSTOM_EVALS[ep.name] = ep.load()
else:
    # Prior to Python 3.10, this returns a dict instead of the selection interface
    if TYPE_CHECKING:
        assert isinstance(all_entry_points, dict)
    for ep in all_entry_points.get(PLUGIN_ENTRY_POINT, []):
        CUSTOM_EVALS[ep.name] = ep.load()

__all__ = [
    "prepareQuery",
    "prepareUpdate",
    "processUpdate",
    "operators",
    "parser",
    "parserutils",
    "CUSTOM_EVALS",
]



"""
SPARQL 1.1 defines path operators for combining/repeating predicates
in triple-patterns.

We overload some Python operators on URIRefs to allow creating path
operators directly in Python.

============ =========================================
Operator     Path
============ =========================================
``p1 / p2``  Path sequence
``p1 | p2``  Path alternative
``p1 * '*'`` chain of 0 or more p's
``p1 * '+'`` chain of 1 or more p's
``p1 * '?'`` 0 or 1 p
``~p1``      p1 inverted, i.e. (s p1 o) <=> (o ~p1 s)
``-p1``      NOT p1, i.e. any property but p1
============ =========================================


These can then be used in property position for ``s,p,o`` triple queries
for any graph method.

See the docs for :mod:`rdflib.paths` for the details.

This example shows how to get the name of friends (i.e values two steps away x knows y, y name z) with a single query.
"""

from pathlib import Path

from rdflib import Graph, URIRef
from rdflib.namespace import FOAF

EXAMPLES_DIR = Path(__file__).parent

if __name__ == "__main__":
    g = Graph()
    g.parse(f"{EXAMPLES_DIR / 'foaf.n3'}")

    tim = URIRef("http://www.w3.org/People/Berners-Lee/card#i")

    print("Timbl knows:")

    for o in g.objects(tim, FOAF.knows / FOAF.name):
        print(o)



"""
Tests for usage of the Store interface from Graph/NamespaceManager.
"""

from __future__ import annotations

import itertools
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
)
from unittest.mock import patch

import pytest

import rdflib.namespace
from rdflib.graph import ConjunctiveGraph, Dataset, Graph
from rdflib.namespace import Namespace
from rdflib.plugins.sparql.sparql import Query
from rdflib.plugins.stores.memory import Memory
from rdflib.query import Result
from rdflib.store import Store
from rdflib.term import Identifier, URIRef, Variable
from test.data import SIMPLE_TRIPLE_GRAPH

if TYPE_CHECKING:
    from _pytest.mark.structures import ParameterSet

NamespaceBindings = Dict[str, URIRef]


def check_ns(graph: Graph, expected_bindings: NamespaceBindings) -> None:
    actual_graph_bindings = list(graph.namespaces())
    logging.info("actual_bindings = %s", actual_graph_bindings)
    assert list(expected_bindings.items()) == actual_graph_bindings
    store: Store = graph.store
    actual_store_bindings = list(store.namespaces())
    assert list(expected_bindings.items()) == actual_store_bindings
    for prefix, uri in expected_bindings.items():
        assert store.prefix(uri) == prefix
        assert store.namespace(prefix) == uri


class MemoryWithoutBindOverride(Memory):
    def bind(self, prefix, namespace) -> None:  # type: ignore[override]
        return super().bind(prefix, namespace, False)


class GraphWithoutBindOverrideFix(Graph):
    def bind(self, prefix, namespace, override=True, replace=False) -> None:
        old_value = rdflib.namespace._with_bind_override_fix
        rdflib.namespace._with_bind_override_fix = False
        try:
            return super().bind(prefix, namespace, override, replace)
        finally:
            rdflib.namespace._with_bind_override_fix = old_value


GraphFactory = Callable[[], Graph]
GraphOperation = Callable[[Graph], None]
GraphOperations = Sequence[GraphOperation]

EGNS = Namespace("http://example.com/namespace#")
EGNS_V0 = EGNS["v0"]
EGNS_V1 = EGNS["v1"]
EGNS_V2 = EGNS["v2"]


def make_graph_store_bind_cases(
    store_type: Type[Store] = Memory,
    graph_type: Type[Graph] = Graph,
) -> Iterable[Union[Tuple[Any, ...], ParameterSet]]:
    """
    Generate test cases for test_graph_store_bind.
    """

    def graph_factory():
        return graph_type(bind_namespaces="none", store=store_type())

    id_prefix = f"{store_type.__name__}-{graph_type.__name__}"

    def _p(
        graph_factory: GraphFactory,
        ops: GraphOperations,
        expected_bindings: NamespaceBindings,
        expected_bindings_overrides: Optional[
            Dict[Tuple[Type[Graph], Type[Store]], NamespaceBindings]
        ] = None,
        *,
        id: Optional[str] = None,
    ):
        if expected_bindings_overrides is not None:
            expected_bindings = expected_bindings_overrides.get(
                (graph_type, store_type), expected_bindings
            )
        if id is not None:
            return pytest.param(graph_factory, ops, expected_bindings, id=id)
        else:
            return (graph_factory, ops, expected_bindings)

    yield from [
        _p(
            graph_factory,
            [
                lambda g: g.bind("eg", EGNS_V0),
            ],
            {"eg": EGNS_V0},
            id=f"{id_prefix}-default-args",
        ),
        # reused-prefix
        _p(
            graph_factory,
            [
                lambda g: g.bind("eg", EGNS_V0),
                lambda g: g.bind("eg", EGNS_V1, override=False),
            ],
            {"eg": EGNS_V0, "eg1": EGNS_V1},
            id=f"{id_prefix}-reused-prefix-override-false-replace-false",
        ),
        _p(
            graph_factory,
            [
                lambda g: g.bind("eg", EGNS_V0),
                lambda g: g.bind("eg", EGNS_V1),
            ],
            {"eg": EGNS_V0, "eg1": EGNS_V1},
            id=f"{id_prefix}-reused-prefix-override-true-replace-false",
        ),
        _p(
            graph_factory,
            [
                lambda g: g.bind("eg", EGNS_V0),
                lambda g: g.bind("eg", EGNS_V1, override=False, replace=True),
            ],
            {"eg": EGNS_V0},
            {(GraphWithoutBindOverrideFix, Memory): {"eg": EGNS_V1}},
            id=f"{id_prefix}-reused-prefix-override-false-replace-true",
        ),
        _p(
            graph_factory,
            [
                lambda g: g.bind("eg", EGNS_V0),
                lambda g: g.bind("eg", EGNS_V1, replace=True),
            ],
            {"eg": EGNS_V1},
            {(Graph, MemoryWithoutBindOverride): {"eg": EGNS_V0}},
            id=f"{id_prefix}-reused-prefix-override-true-replace-true",
        ),
        # reused-namespace
        _p(
            graph_factory,
            [
                lambda g: g.bind("eg", EGNS_V0),
                lambda g: g.bind("egv0", EGNS_V0, override=False),
            ],
            {"eg": EGNS_V0},
            id=f"{id_prefix}-reused-namespace-override-false-replace-false",
        ),
        _p(
            graph_factory,
            [
                lambda g: g.bind("eg", EGNS_V0),
                lambda g: g.bind("egv0", EGNS_V0),
            ],
            {"egv0": EGNS_V0},
            {(Graph, MemoryWithoutBindOverride): {"eg": EGNS_V0}},
            id=f"{id_prefix}-reused-namespace-override-true-replace-false",
        ),
        _p(
            graph_factory,
            [
                lambda g: g.bind("eg", EGNS_V0),
                lambda g: g.bind("egv0", EGNS_V0, override=False, replace=True),
            ],
            {"eg": EGNS_V0},
            id=f"{id_prefix}-reused-namespace-override-false-replace-true",
        ),
        _p(
            graph_factory,
            [
                lambda g: g.bind("eg", EGNS_V0),
                lambda g: g.bind("egv0", EGNS_V0, replace=True),
            ],
            {"egv0": EGNS_V0},
            {(Graph, MemoryWithoutBindOverride): {"eg": EGNS_V0}},
            id=f"{id_prefix}-reused-namespace-override-true-replace-true",
        ),
    ]


@pytest.mark.parametrize(
    ["graph_factory", "ops", "expected_bindings"],
    itertools.chain(
        make_graph_store_bind_cases(),
        make_graph_store_bind_cases(store_type=MemoryWithoutBindOverride),
        make_graph_store_bind_cases(graph_type=GraphWithoutBindOverrideFix),
    ),
)
def test_graph_store_bind(
    graph_factory: GraphFactory,
    ops: GraphOperations,
    expected_bindings: NamespaceBindings,
) -> None:
    """
    The expected sequence of graph operations results in the expected
    namespace bindings.
    """
    graph = graph_factory()
    for op in ops:
        op(graph)
    check_ns(graph, expected_bindings)


@pytest.mark.parametrize(
    ("graph_factory", "query_graph"),
    [
        (Graph, lambda graph: graph.identifier),
        (ConjunctiveGraph, "__UNION__"),
        (Dataset, lambda graph: graph.default_context.identifier),
        (lambda store: Dataset(store=store, default_union=True), "__UNION__"),
    ],
)
def test_query_query_graph(
    graph_factory: Callable[[Store], Graph],
    query_graph: Union[str, Callable[[Graph], str]],
) -> None:
    """
    The `Graph.query` method passes the correct ``queryGraph`` argument
    to stores that have implemented a `Store.query` method.
    """

    mock_result = Result("SELECT")
    mock_result.vars = [Variable("s"), Variable("p"), Variable("o")]
    mock_result.bindings = [
        {
            Variable("s"): URIRef("http://example.org/subject"),
            Variable("p"): URIRef("http://example.org/predicate"),
            Variable("o"): URIRef("http://example.org/object"),
        },
    ]

    query_string = r"FAKE QUERY, NOT USED"
    store = Memory()
    graph = graph_factory(store)

    if callable(query_graph):
        query_graph = query_graph(graph)

    def mock_query(
        query: Union[Query, str],
        initNs: Mapping[str, Any],  # noqa: N803
        initBindings: Mapping[str, Identifier],  # noqa: N803
        queryGraph: str,  # noqa: N803
        **kwargs,
    ) -> Result:
        assert query_string == query
        assert dict(store.namespaces()) == initNs
        assert {} == initBindings
        assert query_graph == queryGraph
        assert {} == kwargs
        return mock_result

    with patch.object(store, "query", wraps=mock_query) as wrapped_query:
        actual_result = graph.query(query_string)
        assert actual_result.type == "SELECT"
        assert list(actual_result) == list(
            SIMPLE_TRIPLE_GRAPH.triples((None, None, None))
        )
        assert wrapped_query.call_count == 1


@pytest.mark.parametrize(
    ("graph_factory", "query_graph"),
    [
        (Graph, lambda graph: graph.identifier),
        (ConjunctiveGraph, "__UNION__"),
        (Dataset, lambda graph: graph.default_context.identifier),
        (lambda store: Dataset(store=store, default_union=True), "__UNION__"),
    ],
)
def test_update_query_graph(
    graph_factory: Callable[[Store], Graph],
    query_graph: Union[str, Callable[[Graph], str]],
) -> None:
    """
    The `Graph.update` method passes the correct ``queryGraph`` argument
    to stores that have implemented a `Store.update` method.
    """

    update_string = r"FAKE UPDATE, NOT USED"
    store = Memory()
    graph = graph_factory(store)

    if callable(query_graph):
        query_graph = query_graph(graph)

    def mock_update(
        query: Union[Query, str],
        initNs: Mapping[str, Any],  # noqa: N803
        initBindings: Mapping[str, Identifier],  # noqa: N803
        queryGraph: str,  # noqa: N803
        **kwargs,
    ) -> None:
        assert update_string == query
        assert dict(store.namespaces()) == initNs
        assert {} == initBindings
        assert query_graph == queryGraph
        assert {} == kwargs

    with patch.object(store, "update", wraps=mock_update) as wrapped_update:
        graph.update(update_string)
        assert wrapped_update.call_count == 1



from rdflib import Literal
from rdflib.namespace import XSD


class TestNormalizedString:
    def test1(self):
        lit2 = Literal("\two\nw", datatype=XSD.normalizedString)
        lit = Literal("\two\nw", datatype=XSD.string)
        assert str(lit) != str(lit2)

    def test2(self):
        lit = Literal(
            "\tBeing a Doctor Is\n\ta Full-Time Job\r", datatype=XSD.normalizedString
        )
        st = Literal(" Being a Doctor Is  a Full-Time Job ", datatype=XSD.string)
        assert Literal.eq(st, lit) is False
        assert str(lit) == str(st)

    def test3(self):
        lit = Literal("hey\nthere", datatype=XSD.normalizedString).n3()
        assert lit == '"hey there"^^<http://www.w3.org/2001/XMLSchema#normalizedString>'

    def test4(self):
        lit = Literal(
            "hey\nthere\ta tab\rcarriage return", datatype=XSD.normalizedString
        )
        expected = Literal("""hey there a tab carriage return""", datatype=XSD.string)
        assert str(lit) == str(expected)



"""Get all merged PRs since last release, save them to a JSON file"""

import json
import urllib.parse
import urllib.request

# https://api.github.com/search/issues?q=repo:rdflib/rdflib+is:pr+merged:%3E=2023-08-02&per_page=300&page=1
LAST_RELEASE_DATE = "2024-10-17"
ISSUES_URL = "https://api.github.com/search/issues"
ITEMS = []
PAGE = 1

# make sequential requests for each page of PRs
while True:
    params = {
        "q": f"repo:rdflib/rdflib+is:pr+merged:>={LAST_RELEASE_DATE}",
        "per_page": 100,
        "page": PAGE,
    }
    query_string = "&".join([f"{k}={v}" for k, v in params.items()])
    url = ISSUES_URL + "?" + query_string

    print(f"Getting {url}")
    with urllib.request.urlopen(url) as response:
        response_text = response.read()
        link_headers = response.info()["link"].split(",") if response.info()["link"] is not None else None

    json_data = json.loads(response_text)
    ITEMS.extend(json_data["items"])

    keep_going = False
    if link_headers is not None:
        for link in link_headers:
            if 'rel="next"' in link:
                # url = link.strip("<").split(">")[0]
                PAGE += 1
                keep_going = True

    if not keep_going:
        break

with open("merged_prs.json", "w") as f:
    json.dump(ITEMS, f, indent=4)

# split interesting and boring PRs into two lists
good_prs = []
boring_prs = []
for pr in sorted(ITEMS, key=lambda k: k["closed_at"], reverse=True):
    matches = ["bump", "pre-commit.ci"]
    if any(x in pr["title"] for x in matches):
        boring_prs.append(
            f"""* {pr['closed_at'][:10]} - {pr['title']}\n  [PR #{pr['number']}]({pr['html_url']})"""
        )
    else:
        good_prs.append(
            f"""* {pr['closed_at'][:10]} - {pr['title']}\n  [PR #{pr['number']}]({pr['html_url']})"""
        )

for pr in good_prs:
    print(pr)

print()
print()

for pr in boring_prs:
    print(pr)



from rdflib.namespace import DefinedNamespace, Namespace
from rdflib.term import URIRef


class PROF(DefinedNamespace):
    """
    Profiles Vocabulary

    This vocabulary is for describing relationships between standards/specifications, profiles of them and
    supporting artifacts such as validating resources.  This model starts with
    [http://dublincore.org/2012/06/14/dcterms#Standard](dct:Standard) entities which can either be Base
    Specifications (a standard not profiling any other Standard) or Profiles (Standards which do profile others).
    Base Specifications or Profiles can have Resource Descriptors associated with them that defines implementing
    rules for the it. Resource Descriptors must indicate the role they play (to guide, to validate etc.) and the
    formalism they adhere to (dct:format) to allow for content negotiation. A vocabulary of Resource Roles are
    provided alongside this vocabulary but that list is extensible.

    Generated from: https://www.w3.org/ns/dx/prof/profilesont.ttl
    Date: 2020-05-26 14:20:03.542924

    """

    # http://www.w3.org/2002/07/owl#Class
    Profile: URIRef  # A named set of constraints on one or more identified base specifications or other profiles, including the identification of any implementing subclasses of datatypes, semantic interpretations, vocabularies, options and parameters of those base specifications necessary to accomplish a particular function.  This definition includes what are often called "application profiles", "metadata application profiles", or "metadata profiles".
    ResourceDescriptor: URIRef  # A resource that defines an aspect - a particular part or feature - of a Profile
    ResourceRole: URIRef  # The role that an Resource plays

    # http://www.w3.org/2002/07/owl#DatatypeProperty
    hasToken: URIRef  # A preferred alternative identifier for the Profile

    # http://www.w3.org/2002/07/owl#ObjectProperty
    hasArtifact: URIRef  # The URL of a downloadable file with particulars such as its format and role indicated by a Resource Descriptor
    hasResource: URIRef  # A resource which describes the nature of an artifact and the role it plays in relation to a profile
    hasRole: URIRef  # The function of the described artifactresource in the expression of the Profile, such as a specification, guidance documentation, SHACL file etc.
    isInheritedFrom: URIRef  # This property indicates a Resource Descriptor described by this Profile’s base specification that is to be considered a Resource Descriptor for this Profile also
    isProfileOf: URIRef  # A Profile is a profile of a dct:Standard (or a Base Specification or another Profile)
    isTransitiveProfileOf: URIRef  # A base specification an Profile conforms to

    _NS = Namespace("http://www.w3.org/ns/dx/prof/")



from io import BytesIO

from rdflib.graph import ConjunctiveGraph, Graph
from rdflib.term import Literal, URIRef


def test_serialize():
    s1 = URIRef("store:1")
    r1 = URIRef("resource:1")
    r2 = URIRef("resource:2")

    label = URIRef("predicate:label")

    g1 = Graph(identifier=s1)
    g1.add((r1, label, Literal("label 1", lang="en")))
    g1.add((r1, label, Literal("label 2")))

    s2 = URIRef("store:2")
    g2 = Graph(identifier=s2)
    g2.add((r2, label, Literal("label 3")))

    g = ConjunctiveGraph()
    for s, p, o in g1.triples((None, None, None)):
        g.addN([(s, p, o, g1)])
    for s, p, o in g2.triples((None, None, None)):
        g.addN([(s, p, o, g2)])
    r3 = URIRef("resource:3")
    g.add((r3, label, Literal(4)))

    r = g.serialize(format="trix", encoding="utf-8")
    g3 = ConjunctiveGraph()

    g3.parse(BytesIO(r), format="trix")

    for q in g3.quads((None, None, None)):
        # TODO: Fix once getGraph/getContext is in conjunctive graph
        if isinstance(q[3].identifier, URIRef):
            tg = Graph(store=g.store, identifier=q[3].identifier)
        else:
            # BNode, this is a bit ugly
            # we cannot match the bnode to the right graph automagically
            # here I know there is only one anonymous graph,
            # and that is the default one, but this is not always the case
            tg = g.default_context

        assert q[0:3] in tg


def test_issue_250():
    """

    https://github.com/RDFLib/rdflib/issues/250

    When I have a ConjunctiveGraph with the default namespace set,
    for example

    import rdflib
    g = rdflib.ConjunctiveGraph()
    g.bind(None, "http://defaultnamespace")

    then the Trix serializer binds the default namespace twice in its XML
    output, once for the Trix namespace and once for the namespace I used:

    print(g.serialize(format='trix').decode('UTF-8'))

    <?xml version="1.0" encoding="utf-8"?>
    <TriX
      xmlns:xml="http://www.w3.org/XML/1998/namespace"
      xmlns="http://defaultnamespace"
      xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
      xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#"
      xmlns="http://www.w3.org/2004/03/trix/trix-1/"
    />

    """

    graph = ConjunctiveGraph()
    graph.bind(None, "http://defaultnamespace")
    sg = graph.serialize(format="trix")
    assert 'xmlns="http://defaultnamespace"' not in sg, sg
    assert 'xmlns="http://www.w3.org/2004/03/trix/trix-1/' in sg, sg



from __future__ import annotations

import logging
from collections import defaultdict
from http.server import BaseHTTPRequestHandler, HTTPServer
from threading import Thread
from types import TracebackType
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ContextManager,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    cast,
)
from unittest.mock import MagicMock, Mock
from urllib.parse import parse_qs, urlparse

from test.utils.http import (
    MethodName,
    MockHTTPRequest,
    MockHTTPResponse,
    apply_headers_to,
    get_random_ip,
)

__all__: List[str] = ["make_spypair", "BaseHTTPServerMock", "ServedBaseHTTPServerMock"]

if TYPE_CHECKING:
    import typing_extensions as te


GenericT = TypeVar("GenericT", bound=Any)


def make_spypair(method: GenericT) -> Tuple[GenericT, Mock]:
    m = MagicMock()

    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        m(*args, **kwargs)
        return method(self, *args, **kwargs)

    setattr(wrapper, "mock", m)
    return cast(GenericT, wrapper), m


RequestDict = Dict[MethodName, List[MockHTTPRequest]]
ResponseDict = Dict[MethodName, List[MockHTTPResponse]]


class BaseHTTPServerMock:
    def __init__(self) -> None:
        self.requests: Dict[MethodName, List[MockHTTPRequest]] = defaultdict(
            lambda: list()
        )
        self.responses: Dict[MethodName, List[MockHTTPResponse]] = defaultdict(
            lambda: list()
        )
        self.mocks: Dict[MethodName, Mock] = {}

        class Handler(BaseHTTPRequestHandler):
            pass

        self.Handler = Handler

        for name in MethodName:
            name_str = name.name
            do_handler, mock = make_spypair(
                self.make_do_handler(name, self.requests, self.responses)
            )
            setattr(self.Handler, f"do_{name_str}", do_handler)
            self.mocks[name] = mock

    @classmethod
    def make_do_handler(
        cls, method_name: MethodName, requests: RequestDict, responses: ResponseDict
    ) -> Callable[[BaseHTTPRequestHandler], None]:
        def do_handler(handler: BaseHTTPRequestHandler) -> None:
            parsed_path = urlparse(handler.path)
            path_query = parse_qs(parsed_path.query)
            body = None
            content_length = handler.headers.get("Content-Length")
            if content_length is not None:
                body = handler.rfile.read(int(content_length))
            request = MockHTTPRequest(
                method_name,
                handler.path,
                parsed_path,
                path_query,
                handler.headers,
                body,
            )
            logging.debug("handling %s request: %s", method_name, request)
            logging.debug("headers %s", request.headers)
            requests[method_name].append(request)

            try:
                response = responses[method_name].pop(0)
            except IndexError as error:
                raise ValueError(f"No response for {method_name} request") from error
            handler.send_response(response.status_code, response.reason_phrase)
            apply_headers_to(response.headers, handler)
            handler.end_headers()

            handler.wfile.write(response.body)
            handler.wfile.flush()
            return

        return do_handler

    def reset(self) -> None:
        self.requests.clear()
        self.responses.clear()
        for name in MethodName:
            self.mocks[name].reset_mock()

    @property
    def call_count(self) -> int:
        return sum(self.mocks[name].call_count for name in MethodName)


class ServedBaseHTTPServerMock(
    BaseHTTPServerMock, ContextManager["ServedBaseHTTPServerMock"]
):
    def __init__(self, host: Optional[str] = "127.0.0.1") -> None:
        super().__init__()
        host = get_random_ip() if host is None else host
        self.server = HTTPServer((host, 0), self.Handler)
        self.server_thread = Thread(target=self.server.serve_forever)
        self.server_thread.daemon = True
        self.server_thread.start()

    def stop(self) -> None:
        self.server.shutdown()
        self.server.socket.close()
        self.server_thread.join()

    @property
    def address_string(self) -> str:
        (host, port) = self.server.server_address
        if isinstance(host, (bytes, bytearray)):
            host = host.decode("utf-8")
        return f"{host}:{port}"

    @property
    def url(self) -> str:
        return f"http://{self.address_string}"

    def __enter__(self) -> ServedBaseHTTPServerMock:
        return self

    def __exit__(
        self,
        __exc_type: Optional[Type[BaseException]],
        __exc_value: Optional[BaseException],
        __traceback: Optional[TracebackType],
    ) -> te.Literal[False]:
        self.stop()
        return False



from rdflib import Graph, URIRef
from rdflib.term import Literal

query_tpl = """
SELECT ?x (MIN(?y_) as ?y) (%s(DISTINCT ?z_) as ?z) {
  VALUES (?x ?y_ ?z_) {
    ("x1" 10 1)
    ("x1" 11 1)
    ("x2" 20 2)
  }
} GROUP BY ?x ORDER BY ?x
"""


def test_group_concat_distinct():
    g = Graph()
    results = g.query(query_tpl % "GROUP_CONCAT")
    results = [[lit.toPython() for lit in line] for line in results]

    # this is the tricky part
    assert results[0][2] == "1", results[0][2]

    # still check the whole result, to be on the safe side
    assert results == [
        ["x1", 10, "1"],
        ["x2", 20, "2"],
    ], results


def test_sum_distinct():
    g = Graph()
    results = g.query(query_tpl % "SUM")
    results = [[lit.toPython() for lit in line] for line in results]

    # this is the tricky part
    assert results[0][2] == 1, results[0][2]

    # still check the whole result, to be on the safe side
    assert results == [
        ["x1", 10, 1],
        ["x2", 20, 2],
    ], results


def test_avg_distinct():
    g = Graph()
    results = g.query(
        """
        SELECT ?x (MIN(?y_) as ?y) (AVG(DISTINCT ?z_) as ?z) {
          VALUES (?x ?y_ ?z_) {
            ("x1" 10 1)
            ("x1" 11 1)
            ("x1" 12 3)
            ("x2" 20 2)
          }
       } GROUP BY ?x ORDER BY ?x
    """
    )
    results = [[lit.toPython() for lit in line] for line in results]

    # this is the tricky part
    assert results[0][2] == 2, results[0][2]

    # still check the whole result, to be on the safe side
    assert results == [
        ["x1", 10, 2],
        ["x2", 20, 2],
    ], results


def test_count_distinct():
    g = Graph()

    g.parse(
        format="turtle",
        publicID="http://example.org/",
        data="""
    @prefix : <> .

    <#a>
      :knows <#b>, <#c> ;
      :age 42 .

    <#b>
      :knows <#a>, <#c> ;
      :age 36 .

    <#c>
      :knows <#b>, <#c> ;
      :age 20 .

    """,
    )

    # Query 1: people knowing someone younger
    results = g.query(
        """
    PREFIX : <http://example.org/>

    SELECT DISTINCT ?x {
      ?x :age ?ax ; :knows [ :age ?ay ].
      FILTER( ?ax > ?ay )
    }
    """
    )
    assert len(results) == 2

    # nQuery 2: count people knowing someone younger
    results = g.query(
        """
    PREFIX : <http://example.org/>

    SELECT (COUNT(DISTINCT ?x) as ?cx) {
      ?x :age ?ax ; :knows [ :age ?ay ].
      FILTER( ?ax > ?ay )
    }
    """
    )
    assert list(results)[0][0].toPython() == 2


def test_count_optional_values():
    """Problematic query because ?inst may be not bound.
    So when counting over not bound variables it throws a NotBoundError.
    """
    g = Graph()
    g.bind("ex", "http://example.com/")
    g.parse(
        format="ttl",
        data="""@prefix ex: <http://example.com/>.
            ex:1 a ex:a;
                ex:d ex:b.
            ex:2 a ex:a;
                ex:d ex:c;
                ex:d ex:b.
            ex:3 a ex:a.
    """,
    )

    query = """
    SELECT DISTINCT ?x (COUNT(DISTINCT ?inst) as ?cnt)
    WHERE {
        ?x a ex:a
        OPTIONAL {
            VALUES ?inst {ex:b ex:c}.
            ?x ex:d ?inst.
        }
    }  GROUP BY ?x
    """
    results = dict(g.query(query))
    assert results == {
        URIRef("http://example.com/1"): Literal(1),
        URIRef("http://example.com/2"): Literal(2),
        URIRef("http://example.com/3"): Literal(0),
    }



"""
Issue 923: split charset off of Content-Type before looking up Result-parsing plugin.
"""

from io import StringIO

from rdflib.query import Result

RESULT_SOURCE = """\
{
  "head" : {
    "vars" : [ "subject", "predicate", "object", "context" ]
  },
  "results" : {
    "bindings" : [ {
      "subject" : {
        "type" : "bnode",
        "value" : "service"
      },
      "predicate" : {
        "type" : "uri",
        "value" : "http://www.w3.org/1999/02/22-rdf-syntax-ns#type"
      },
      "object" : {
        "type" : "uri",
        "value" : "http://www.w3.org/ns/sparql-service-description#Service"
      }
    }]
  }
}
"""


def test_issue_923():
    with StringIO(RESULT_SOURCE) as result_source:
        Result.parse(
            source=result_source,
            content_type="application/sparql-results+json;charset=utf-8",
        )



from __future__ import annotations

import abc
import contextlib
import logging
from collections.abc import Iterable as IterableABC
from dataclasses import dataclass
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterable,
    NoReturn,
    Optional,
    Pattern,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

import pytest
from pytest import ExceptionInfo

AnyT = TypeVar("AnyT")

OutcomePrimitive = Union[
    AnyT, Callable[[AnyT], None], "OutcomeChecker[AnyT]", Type[Exception], Exception
]

OutcomePrimitives = Union[
    Iterable[Union[AnyT, Callable[[AnyT], None], "OutcomeChecker[AnyT]"]],
    OutcomePrimitive,
]


class OutcomeChecker(abc.ABC, Generic[AnyT]):
    """
    Validates expected outcomes for tests.

    Useful for parameterized test that can result in values or
    exceptions.
    """

    @abc.abstractmethod
    def check(self, actual: AnyT) -> None:
        """
        Check the actual outcome against the expectation.

        This should run inside the checker's context.

        :param outcome: The actual outcome of the test.
        :raises AssertionError: If the outcome does not match the
            expectation.
        :raises RuntimeError: If this method is called when no outcome
            is expected.
        """
        ...

    @contextlib.contextmanager
    @abc.abstractmethod
    def context(self) -> Generator[Optional[ExceptionInfo[Exception]], None, None]:
        """
        The context in which the test code should run.

        This is necessary for checking exception outcomes.

        :return: A context manager that yields the exception info for
            any exceptions that were raised in this context.
        :raises AssertionError: If the test does not raise an exception
            when one is expected, or if the exception does not match the
            expectation.
        """
        ...

    @classmethod
    def from_primitive(
        cls,
        primitive: OutcomePrimitive[AnyT],
    ) -> OutcomeChecker[AnyT]:
        checker = cls._from_special(primitive)
        if checker is not None:
            return checker
        return ValueChecker(cast(AnyT, primitive))

    @classmethod
    def _from_special(
        cls,
        primitive: Union[
            AnyT,
            Callable[[AnyT], None],
            OutcomeChecker[AnyT],
            Type[Exception],
            Exception,
        ],
    ) -> Optional[OutcomeChecker[AnyT]]:
        if isinstance(primitive, OutcomeChecker):
            return primitive
        if isinstance(primitive, type) and issubclass(primitive, Exception):
            return ExceptionChecker(primitive)
        if isinstance(primitive, Exception):
            return ExceptionChecker(type(primitive), match=primitive.args[0])
        if callable(primitive):
            return CallableChecker(cast(Callable[[AnyT], None], primitive))
        return None

    @classmethod
    def from_primitives(
        cls,
        primitives: OutcomePrimitives[AnyT],
    ) -> OutcomeChecker[AnyT]:
        checker = cls._from_special(primitives)  # type: ignore[arg-type]
        if checker is not None:
            return checker
        if isinstance(primitives, IterableABC) and not isinstance(
            primitives, (str, bytes)
        ):
            primitives = iter(primitives)
            return AggregateChecker([cls.from_primitive(p) for p in primitives])
        return ValueChecker(cast(AnyT, primitives))


@dataclass(frozen=True)
class NoExceptionChecker(OutcomeChecker[AnyT]):
    """
    Base class for checkers that do not expect exceptions.
    """

    @contextlib.contextmanager
    def context(self) -> Generator[None, None, None]:
        yield None


@dataclass(frozen=True)
class AggregateChecker(NoExceptionChecker[AnyT]):
    """
    Validates that the outcome matches all of the given checkers.
    """

    checkers: Sequence[OutcomeChecker[AnyT]]

    def check(self, actual: AnyT) -> None:
        for checker in self.checkers:
            if isinstance(checker, ExceptionChecker):
                raise ValueError(
                    "AggregateChecker should never contain ExceptionChecker"
                )
            checker.check(actual)


@dataclass(frozen=True)
class ValueChecker(NoExceptionChecker[AnyT]):
    """
    Validates that the outcome is a specific value.

    :param value: The expected value.
    """

    expected: AnyT

    def check(self, actual: AnyT) -> None:
        assert self.expected == actual


@dataclass(frozen=True)
class CallableChecker(NoExceptionChecker[AnyT]):
    """
    Validates the outcome with a callable.

    :param callable: The callable that will be called with the outcome
        to validate it.
    """

    callable: Callable[[AnyT], None]

    def check(self, actual: AnyT) -> None:
        self.callable(actual)


@dataclass(frozen=True)
class ExceptionChecker(OutcomeChecker[AnyT]):
    """
    Validates that the outcome is a specific exception.

    :param type: The expected exception type.
    :param match: A regular expression or string that the exception
        message must match.
    :param attributes: A dictionary of attributes that the exception
        must have and their expected values.
    """

    type: Type[Exception]
    match: Optional[Union[Pattern[str], str]] = None
    attributes: Optional[Dict[str, Any]] = None

    def check(self, actual: AnyT) -> NoReturn:
        raise RuntimeError("ExceptionResult.check_result should never be called")

    def _check_attributes(self, exception: Exception) -> None:
        if self.attributes is not None:
            for key, value in self.attributes.items():
                logging.debug("checking exception attribute %s=%r", key, value)
                assert hasattr(exception, key)
                assert getattr(exception, key) == value

    @contextlib.contextmanager
    def context(self) -> Generator[ExceptionInfo[Exception], None, None]:
        with pytest.raises(self.type, match=self.match) as catcher:
            yield catcher
        self._check_attributes(catcher.value)



import pytest

from rdflib import ConjunctiveGraph, Graph, Literal
from rdflib.namespace import OWL, Namespace, NamespaceManager
from rdflib.plugins.stores.memory import Memory
from rdflib.term import URIRef
from test.data import CONTEXT1, CONTEXT2, TAREK

foaf1_uri = URIRef("http://xmlns.com/foaf/0.1/")
foaf2_uri = URIRef("http://xmlns.com/foaf/2.0/")

FOAF1 = Namespace(foaf1_uri)
FOAF2 = Namespace(foaf2_uri)


def test_binding_replace():
    # The confusion here is in the two arguments “override” and “replace” and
    # how they serve two different purposes --- changing a prefix already bound
    # to a namespace versus changing a namespace already bound to a prefix.

    g = Graph(bind_namespaces="none")
    assert len(list(g.namespaces())) == 0

    # 1. Changing the namespace of an existing namespace=>prefix binding:

    # Say they release FOAF 2.0 and you want to change the binding of
    # `http://xmlns.com/foaf/2.0/` to `foaf`.

    # Up to now you had `http://xmlns.com/foaf/0.1/=>foaf` and `http://xmlns.com/foaf/2.0/=>foaf2`

    # We'll just set up those two bindings ...
    g.bind("foaf", FOAF1)
    g.bind("foaf2", FOAF2)
    assert len(list(g.namespaces())) == 2
    assert list(g.namespaces()) == [("foaf", foaf1_uri), ("foaf2", foaf2_uri)]

    # Now you want to "upgrade" to FOAF2=>foaf and try the obvious:
    g.bind("foaf", FOAF2)

    # But that doesn't happen, instead a different prefix, `foaf1` is bound:
    assert list(g.namespaces()) == [("foaf", foaf1_uri), ("foaf1", foaf2_uri)]

    # The rationale behind this behaviour is that the prefix "foaf" was already
    # bound to the FOAF1 namespace, so a different prefix for the FOAF2 namespace
    # was automatically created.

    # You can achieve the intended result by setting `replace` to `True`:
    g.bind("foaf", FOAF2, replace=True)

    # Because the FOAF2 namespace was rebound to a different prefix, the old
    # "foaf2" prefix has gone and because the "foaf" prefix was rebound to a
    # different namespace, the old FOAF1 namespace has gone, leaving just:

    assert list(g.namespaces()) == [("foaf", foaf2_uri)]

    # Now, if you wish, you can re-bind the FOAF1.0 namespace to a prefix
    # of your choice
    g.bind("oldfoaf", FOAF1)

    assert list(g.namespaces()) == [
        ("foaf", foaf2_uri),  # Should be present but has been removed.
        ("oldfoaf", foaf1_uri),
    ]

    # 2. Changing the prefix of an existing namespace=>prefix binding:

    # The namespace manager preserves the existing bindings from any
    # subsequent unwanted programmatic rebindings such as:
    g.bind("foaf", FOAF1)

    # Which, as things stand, results in:

    assert list(g.namespaces()) == [("foaf", foaf2_uri), ("foaf1", foaf1_uri)]

    # In which the attempted change from `oldfoaf` to (the already
    # bound-to-a different-namespace `foaf`) was intercepted and
    # changed to a non-clashing prefix of `foaf1`

    # So, after undoing the above unwanted rebinding ..
    g.bind("oldfoaf", FOAF1, replace=True)

    # The bindings are again as desired
    assert list(g.namespaces()) == [("foaf", foaf2_uri), ("oldfoaf", foaf1_uri)]

    # Next time through, set override to False
    g.bind("foaf", FOAF1, override=False)

    # And the bindings will remain as desired
    assert list(g.namespaces()) == [("foaf", foaf2_uri), ("oldfoaf", foaf1_uri)]

    # 3. Parsing data with prefix=>namespace bindings
    # Let's see the situation regarding namespace bindings
    # in parsed data - it can be a bit confusing ...

    # Starting with a very likely example where `foaf` is a
    # prefix for `FOAF1`

    data = """\
    @prefix foaf: <http://xmlns.com/foaf/0.1/> .

    <urn:example:a> foaf:name "Bob" ."""

    g.parse(data=data, format="n3")

    # The FOAF2 namespace is already bound to `foaf` so a
    # non-clashing prefix of `foaf1` is rebound to FOAF1 in
    # place of the existing `oldfoaf` prefix

    assert list(g.namespaces()) == [("foaf", foaf2_uri), ("foaf1", foaf1_uri)]

    # Yes, namespace bindings in parsed data replace existing
    # bindings (i.e. `oldfoaf` was replaced by `foaf1`), just
    # live with it ...

    # A different example of the same principle where `foaf2`
    # replaces the exsting `foaf`

    data = """\
    @prefix foaf2: <http://xmlns.com/foaf/2.0/> .

    <urn:example:b> foaf2:name "Alice" ."""

    g.parse(data=data, format="n3")

    # The already-bound namespace of `foaf=>FOAF2` is replaced
    # by the `foaf2=>FOAF2` binding specified in the N3 source

    assert list(g.namespaces()) == [
        ("foaf1", foaf1_uri),
        ("foaf2", foaf2_uri),
    ]

    # Where a prefix-namespace binding in the data does
    # not clash with the existing binding ...

    data = """\
    @prefix foaf1: <http://xmlns.com/foaf/0.1/> .

    <urn:example:a> foaf1:name "Bob" ."""

    g.parse(data=data, format="n3")

    # The prefix `foaf1`, already bound to FOAF1 is
    # used

    assert list(g.namespaces()) == [
        ("foaf1", foaf1_uri),
        ("foaf2", foaf2_uri),
    ]

    # Otherwise, the existing prefix binding is replaced

    data = """\
    @prefix foaf: <http://xmlns.com/foaf/0.1/> .

    <urn:example:b> foaf:name "Alice" ."""

    g.parse(data=data, format="n3")

    assert list(g.namespaces()) == [
        ("foaf2", foaf2_uri),
        ("foaf", foaf1_uri),
    ]

    # Prefixes are bound to namespaces which in turn have a URIRef
    # representation - so a different prefix=>namespace binding
    # means a different predicate and thus a different statement:

    expected = """
        @prefix foaf: <http://xmlns.com/foaf/0.1/> .
        @prefix foaf2: <http://xmlns.com/foaf/2.0/> .

        <urn:example:a> foaf:name "Bob" .

        <urn:example:b> foaf:name "Alice" ;
            foaf2:name "Alice" .

        """

    s = g.serialize(format="n3")

    for l in expected.split():  # noqa: E741
        assert l in s


def test_prefix_alias_disallowed():
    # CANNOT BIND A PREFIX ALIASED TO AN EXISTING BOUND NAMESPACE

    g = Graph(bind_namespaces="none")
    g.bind("owl", OWL)
    assert ("owl", URIRef("http://www.w3.org/2002/07/owl#")) in list(g.namespaces())

    g.bind("wol", OWL, override=False)
    assert ("owl", URIRef("http://www.w3.org/2002/07/owl#")) in list(g.namespaces())
    assert ("wol", URIRef("http://www.w3.org/2002/07/owl#")) not in list(g.namespaces())


def test_rebind_prefix_succeeds():
    # CAN REPLACE A PREFIX-NAMESPACE BINDING

    g = Graph(bind_namespaces="none")
    g.bind("owl", OWL)
    assert ("owl", URIRef("http://www.w3.org/2002/07/owl#")) in list(g.namespaces())

    g.bind("wol", OWL)
    assert ("wol", URIRef("http://www.w3.org/2002/07/owl#")) in list(g.namespaces())
    assert ("owl", URIRef("http://www.w3.org/2002/07/owl#")) not in list(g.namespaces())


def test_parse_rebinds_prefix():
    # PARSED PREFIX-NAMESPACE BINDINGS REPLACE EXISTING BINDINGS

    data = """@prefix friend-of-a-friend: <http://xmlns.com/foaf/0.1/> .

    <urn:example:a> friend-of-a-friend:name "Bob" .

    """

    # Use full set of namespace bindings, including foaf
    g = Graph(bind_namespaces="none")
    g.bind("foaf", FOAF1)
    assert ("foaf", foaf1_uri) in list(g.namespaces())

    g.parse(data=data, format="n3")

    # foaf no longer in set of namespace bindings
    assert ("foaf", foaf1_uri) not in list(g.namespaces())
    assert ("friend-of-a-friend", foaf1_uri) in list(g.namespaces())


@pytest.mark.xfail(
    reason="""
    Automatic handling of unknown predicates not automatically registered with namespace manager

    NOTE: This is not a bug, but more of a feature request.
    """
)
def test_automatic_handling_of_unknown_predicates():
    # AUTOMATIC HANDLING OF UNKNOWN PREDICATES

    """
    Automatic handling of unknown predicates
    -----------------------------------------

    As a programming convenience, a namespace binding is automatically
    created when :class:`rdflib.term.URIRef` predicates are added to the graph.
    """

    g = Graph(bind_namespaces="none")

    g.add((TAREK, URIRef("http://xmlns.com/foaf/0.1/name"), Literal("Tarek")))

    assert len(list(g.namespaces())) > 0


def test_automatic_handling_of_unknown_predicates_only_effected_after_serialization():
    g = Graph(bind_namespaces="none")

    g.add((TAREK, URIRef("http://xmlns.com/foaf/0.1/name"), Literal("Tarek")))

    assert "@prefix ns1: <http://xmlns.com/foaf/0.1/> ." in g.serialize(format="n3")

    assert len(list(g.namespaces())) > 0
    assert ("ns1", URIRef("http://xmlns.com/foaf/0.1/")) in list(g.namespaces())


def test_multigraph_bindings():
    data = """@prefix friend-of-a-friend: <http://xmlns.com/foaf/0.1/> .

    <urn:example:a> friend-of-a-friend:name "Bob" .

    """

    store = Memory()

    g1 = Graph(store, identifier=CONTEXT1, bind_namespaces="none")

    g1.bind("foaf", FOAF1)
    assert list(g1.namespaces()) == [("foaf", foaf1_uri)]
    assert list(store.namespaces()) == [("foaf", foaf1_uri)]

    g1.add((TAREK, FOAF1.name, Literal("tarek")))

    assert list(store.namespaces()) == [("foaf", foaf1_uri)]

    g2 = Graph(store, identifier=CONTEXT2, bind_namespaces="none")
    g2.parse(data=data, format="n3")

    # The parser-caused rebind is in the underlying store and all objects
    # that use the store see the changed binding:
    assert list(store.namespaces()) == [("friend-of-a-friend", foaf1_uri)]
    assert list(g1.namespaces()) == [("friend-of-a-friend", foaf1_uri)]

    # Including newly-created objects that use the store
    cg = ConjunctiveGraph(store=store)
    cg.namespace_manager = NamespaceManager(cg, bind_namespaces="core")

    assert ("foaf", foaf1_uri) not in list(cg.namespaces())
    assert ("friend-of-a-friend", foaf1_uri) in list(cg.namespaces())

    assert len(list(g1.namespaces())) == 6
    assert len(list(g2.namespaces())) == 6
    assert len(list(cg.namespaces())) == 6
    assert len(list(store.namespaces())) == 6

    cg.store.add_graph(g1)
    cg.store.add_graph(g2)

    assert "@prefix friend-of-a-friend: <http://xmlns.com/foaf/0.1/>" in cg.serialize(
        format="n3"
    )

    # In the notation3 format, the statement asserting TAREK's name
    # now references the changed prefix:
    assert '<urn:example:tarek> friend-of-a-friend:name "tarek" .' in cg.serialize(
        format="n3"
    )

    # Add foaf2 binding if not already bound
    cg.bind("foaf2", FOAF2, override=False)
    assert ("foaf2", foaf2_uri) in list(cg.namespaces())

    # Impose foaf binding ... if not already bound
    cg.bind("foaf", FOAF1, override=False)
    assert ("foaf", foaf1_uri) not in list(cg.namespaces())


def test_new_namespace_new_prefix():
    g = Graph(bind_namespaces="none")
    g.bind("foaf", FOAF1)
    assert list(g.namespaces()) == [("foaf", foaf1_uri)]


def test_change_prefix_override_true():
    g = Graph(bind_namespaces="none")

    g.bind("foaf", FOAF1)
    assert list(g.namespaces()) == [("foaf", foaf1_uri)]

    g.bind("oldfoaf", FOAF1)
    # Changed
    assert list(g.namespaces()) == [("oldfoaf", foaf1_uri)]


def test_change_prefix_override_false():
    g = Graph(bind_namespaces="none")

    g.bind("foaf", FOAF1)
    assert list(g.namespaces()) == [("foaf", foaf1_uri)]

    g.bind("oldfoaf", FOAF1, override=False)
    # No change
    assert list(g.namespaces()) == [("foaf", foaf1_uri)]


def test_change_namespace_override_true():
    g = Graph(bind_namespaces="none")

    g.bind("foaf", FOAF1)
    assert list(g.namespaces()) == [("foaf", foaf1_uri)]

    g.bind("foaf", FOAF2)
    # Different prefix used
    assert list(g.namespaces()) == [("foaf", foaf1_uri), ("foaf1", foaf2_uri)]


def test_change_namespace_override_false():
    g = Graph(bind_namespaces="none")

    g.bind("foaf", FOAF1)
    assert list(g.namespaces()) == [("foaf", foaf1_uri)]

    # Different namespace so override is irrelevant in this case
    g.bind("foaf", FOAF2, override=False)
    # Different prefix used
    assert list(g.namespaces()) == [("foaf", foaf1_uri), ("foaf1", foaf2_uri)]


def test_new_namespace_override_false():
    g = Graph(bind_namespaces="none")

    g.bind("foaf", FOAF2)
    assert list(g.namespaces()) == [("foaf", foaf2_uri)]

    # Namespace not already bound so override is irrelevant in this case
    g.bind("owl", OWL, override=False)
    # Given prefix used
    assert list(g.namespaces()) == [
        ("foaf", foaf2_uri),
        ("owl", URIRef("http://www.w3.org/2002/07/owl#")),
    ]


def test_change_namespace_and_prefix():
    # A more extensive illustration of attempted rebinds of
    # `foaf` being intercepted and changed to a non-clashing
    # prefix of `foafN` (where N is an incrementing integer) ...

    g = Graph(bind_namespaces="none")

    g.bind("foaf", FOAF1)
    assert list(g.namespaces()) == [("foaf", foaf1_uri)]

    g.bind("foaf", FOAF2, replace=True)
    assert list(g.namespaces()) == [("foaf", foaf2_uri)]

    g.bind("foaf1", FOAF1)

    assert list(g.namespaces()) == [("foaf", foaf2_uri), ("foaf1", foaf1_uri)]

    foaf3_uri = URIRef("http://xmlns.com/foaf/3.0/")
    FOAF3 = Namespace("http://xmlns.com/foaf/3.0/")  # noqa: N806

    g.bind("foaf", FOAF3)

    assert list(g.namespaces()) == [
        ("foaf", foaf2_uri),
        ("foaf1", foaf1_uri),
        ("foaf2", foaf3_uri),
    ]

    foaf4_uri = URIRef("http://xmlns.com/foaf/4.0/")
    FOAF4 = Namespace("http://xmlns.com/foaf/4.0/")  # noqa: N806

    g.bind("foaf", FOAF4)

    assert list(g.namespaces()) == [
        ("foaf", foaf2_uri),
        ("foaf1", foaf1_uri),
        ("foaf2", foaf3_uri),
        ("foaf3", foaf4_uri),
    ]
